"""An HDF5-based indexer for local file systems."""

from __future__ import annotations

import abc
import json
import os
import time
import warnings
from contextlib import suppress
from functools import cache
from pathlib import Path

import pandas as pd
import pooch
from typing_extensions import Self

import dascore as dc
from dascore.constants import ONE_SECOND_IN_NS, PROGRESS_LEVELS
from dascore.exceptions import InvalidIndexVersionError
from dascore.utils.hdf5 import HDFPatchIndexManager
from dascore.utils.misc import iterate
from dascore.utils.pd import filter_df
from dascore.utils.time import get_max_min_times, to_timedelta64

# supported read_hdf5 kwargs
READ_HDF5_KWARGS = frozenset(
    {"columns", "where", "mode", "errors", "start", "stop", "key", "chunksize"}
)


@cache
def _get_index_map(cache_path) -> dict:
    """
    Get a dict of index locations.

    Note: this is purposefully mutable; handle with care.
    """
    path = Path(cache_path)
    out = {}
    successful_read = True
    if path.exists():
        try:
            with path.open("r") as fi:
                out = json.load(fi)
        # On rare occasions, the file can become corrupt. See #508.
        except (OSError, json.JSONDecodeError):
            successful_read = False
    if not isinstance(out, dict) or not successful_read:
        out = {}
        with suppress(FileNotFoundError, PermissionError):
            path.unlink(missing_ok=True)
    return out


def _update_index_map(updates, cache_path) -> dict:
    """Update index map to track new index."""
    data = _get_index_map(cache_path=cache_path)
    data.update(updates)
    Path(cache_path).parent.mkdir(exist_ok=True, parents=True)
    with open(cache_path, "w") as fi:
        json.dump(data, fi)
    return data


def _directory_writable(path):
    """Return True if the directory is writable else False."""
    name = "._dascore_write_test_delete_me"
    path = Path(path) / name
    path.parent.mkdir(exist_ok=True, parents=True)
    try:
        open(path, "w").close()
    except (PermissionError, IsADirectoryError):
        return False
    else:
        os.remove(path)
    return True


class AbstractIndexer:
    """
    A base class for indexers.

    This is primarily here for a place-holder.
    """

    path: Path

    @abc.abstractmethod
    def update(self) -> Self:
        """
        Updates the contents of the Indexer.

        Resets any previous selection.
        """


class DirectoryIndexer(AbstractIndexer):
    """
    A class for indexing a directory of dascore-readable files.

    This works by crawling the directory, getting a summary about the data it
    contains, then creating a small HDF index file which can be queried later
    on.

    Parameters
    ----------
    path
        The path to a directory containing DAS files.
    index_path
        The path to the index. By default, the index will be created on the
        top level of the data directory. If another index is
    """

    ext = ""
    _namespace = ""
    _index_name = ".dascore_index.h5"  # name of index file

    # A user-specific file which tracks where indices are stored if not on
    # the same directory as the path to index.
    index_map_path = pooch.os_cache("dascore") / "caches" / "cache_paths.json"

    def __init__(self, path: str | Path, cache_size: int = 5, index_path=None):
        self.max_size = cache_size
        self.path = Path(path).absolute()
        self.index_path = Path(self._find_index_file(self.path, index_path))
        self._current_index = 0
        self._index_table = HDFPatchIndexManager(
            self.index_path,
            self._namespace,
        )
        self.cache = pd.DataFrame(
            index=range(cache_size), columns="t1 t2 kwargs cindex".split()
        )

    def _find_index_file(self, data_path, index_path=None):
        """Find the path to the index file."""
        data_path = Path(data_path).absolute()
        # user specified index path
        if index_path:
            update = {str(data_path): str(Path(index_path).absolute())}
            _update_index_map(update, cache_path=str(self.index_map_path))
            return index_path
        # see if expected path is in data path
        expected_path = data_path / self._index_name
        with suppress(PermissionError):
            if expected_path.exists():
                return expected_path
        # else load path map and see if it knows where the index is.
        path_map = _get_index_map(cache_path=str(self.index_map_path))
        if out := path_map.get(str(data_path)):
            return out
        # if not, set the path to either the data path, if writable,
        # else the dascore cache
        if not _directory_writable(data_path):
            new_path = "_dascore_index_" + str(abs(hash(data_path))) + ".h5"
            index_path = self.index_map_path.parent / new_path
            update = {str(data_path): str(index_path.absolute())}
            _update_index_map(update, cache_path=str(self.index_map_path))
        else:
            index_path = data_path / self._index_name
        return index_path

    def get_contents(self, buffer=ONE_SECOND_IN_NS, **kwargs) -> pd.DataFrame:
        """
        Get contents of directory with specific query params.

        Parameters
        ----------
        buffer
            A buffer to ensure enough info is returned from hdf index.
        kwargs
            Used to query contents.
        """
        # create index if it doesn't exist
        if not self.index_path.exists():
            self.update()
        # if the index still doesn't exist there are no readable files, return
        # empty df.
        if not self.index_path.exists():
            return pd.DataFrame(columns=self._index_table.index_columns)
        time_min, time_max = get_max_min_times(kwargs.pop("time", None))
        hdf5_kwargs, kwargs = self._separate_hdf5_kwargs(kwargs)
        buffer = to_timedelta64(buffer)
        # find out if the query falls within one cached times
        con1 = self.cache.t1 <= time_min
        con2 = self.cache.t2 >= time_max
        con3 = self.cache.kwargs == self._kwargs_to_str(kwargs)
        cached_index = self.cache[con1 & con2 & con3]
        if not len(cached_index):  # query is not cached get it from hdf5 file
            index = self._index_table.get_index(
                time_min=time_min,
                time_max=time_max,
                **hdf5_kwargs,
            )
            self._set_cache(index, time_min, time_max, hdf5_kwargs)
        else:
            index = cached_index.iloc[0]["cindex"]
        # trim down index
        con1 = index["time_min"] >= (time_max + buffer)
        con2 = index["time_max"] <= (time_min - buffer)
        pre_filter_df = index[~(con1 | con2)]
        out = pre_filter_df[
            filter_df(
                pre_filter_df,
                time=(time_min, time_max),
                ignore_bad_kwargs=True,
                **kwargs,
            )
        ]
        return out

    def __str__(self):
        """Rep. indexer as a string."""
        msg = f"{self.__class__.__name__} managing: {self.path}"
        return msg

    __repr__ = __str__

    __call__ = get_contents

    def _separate_hdf5_kwargs(self, kwargs):
        """Ensure kwargs are supported."""
        kdf_kwargs = {i: v for i, v in kwargs.items() if i in READ_HDF5_KWARGS}
        kwargs = {i: v for i, v in kwargs.items() if i not in READ_HDF5_KWARGS}
        return kdf_kwargs, kwargs

    def _set_cache(self, index, starttime, endtime, kwargs):
        """Cache the current index."""
        ser = pd.Series(
            {
                "t1": starttime,
                "t2": endtime,
                "cindex": index,
                "kwargs": self._kwargs_to_str(kwargs),
            }
        )
        self.cache.loc[self._get_next_index()] = ser

    def clear_cache(self):
        """Removes all cached dataframes."""
        self.cache = pd.DataFrame(
            index=range(self.max_size), columns="t1 t2 kwargs cindex".split()
        )

    def _get_next_index(self):
        """
        Get the next index value on cache.
        Note we can't use itertools.cycle here because it cant be pickled.
        """
        if self._current_index == len(self.cache.index) - 1:
            self._current_index = 0
        else:
            self._current_index += 1
        return self.cache.index[self._current_index]

    def _kwargs_to_str(self, kwargs):
        """Convert kwargs to a string."""
        keys = sorted(list(kwargs.keys()))
        out = str([(item, kwargs[item]) for item in keys])
        return out

    def _get_mtime(self, only_new=True):
        """Return an iterator of potential un-indexed files."""
        # get mtime, subtract a bit to avoid odd bugs
        mtime = None
        # getting last updated might need the db so only call once.
        last_updated = self._index_table.last_updated_timestamp if only_new else None
        if last_updated is not None and only_new:
            mtime = last_updated - 0.001
        # get paths to iterate
        return mtime

    def _get_paths(self, paths):
        path = self.path
        if paths is None:
            paths = path
        else:
            paths = [
                f"{path}/{x}" if str(path) not in str(x) else str(x)
                for x in iterate(paths)
            ]
        return paths

    def _enforce_min_version(self):
        """Ensure the minimum version is met, else delete index file."""
        try:
            self._index_table.validate_version()
        except InvalidIndexVersionError:
            msg = (
                f"The index file at {self.path} is not compatible with this"
                f" version of DASCore ({dc.__last_version__}). "
                f"Recreating the index now."
            )
            warnings.warn(msg, UserWarning)
            os.remove(self.index_path)
            self.update()

    def get_index_metadata(self):
        """Return a dict of metadata about the index."""
        self.update()
        up_time = dc.to_datetime64(self._index_table.last_updated_timestamp)
        out = {
            "index_version": self._index_table._index_version,
            "last_update": up_time,
        }
        return out

    def update(self, paths=None, progress: PROGRESS_LEVELS = "standard") -> Self:
        """
        Updates the contents of the Indexer.

        Also resets any previous selection.

        Parameters
        ----------
        paths
            A sequence of paths to limit the updates, if None, index all
            the contents of directory.
        progress
            The type of progress bar to use. None disables progress bar and
            "basic" is best for low latency scenarios.
        """
        self._enforce_min_version()  # delete index if schema has changed
        update_time = time.time()
        timestamp = self._get_mtime(only_new=True)
        paths = self._get_paths(paths)
        df = dc.scan_to_df(
            path=paths,
            timestamp=timestamp,
            progress=progress,
            ext=self.ext,
        )
        # Put contents found into database.
        if not df.empty:
            # Some users were surprised the spool wasn't sorted. We still cant
            # guarantee all spools will be sorted but we can make sure most are
            # by sorting the contents before dumping to index.
            if "time_min" in df.columns:
                df = df.sort_values("time_min").reset_index(drop=True)
            # ensure the base path is not in the path column
            assert "path" in set(df.columns), f"{df} has no path column"
            self._index_table.write_update(df, update_time, base_path=self.path)
            # clear cache out when new traces are added
            self.clear_cache()
        return self



"""Tests for decimation."""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

import dascore as dc
from dascore.compat import random_state
from dascore.exceptions import FilterValueError
from dascore.units import Hz, m, s
from dascore.utils.patch import get_start_stop_step


class TestInterpolate:
    """Tests for interpolating data along an axis in patch."""

    def test_interp_upsample_distance(self, random_patch):
        """Ensure interpolation between distance works."""
        start, stop, step = get_start_stop_step(random_patch, "distance")
        axis = random_patch.dims.index("distance")
        new_sampling = step / 2.2
        new_coord = np.arange(start, stop, new_sampling)
        out = random_patch.interpolate(distance=new_coord)
        assert out.data.shape[axis] == len(new_coord)
        assert np.allclose(out.coords.get_array("distance"), new_coord)

    def test_interp_down_sample_distance(self, random_patch):
        """Ensure interp can be used to downsample data."""
        start, stop, step = get_start_stop_step(random_patch, "distance")
        axis = random_patch.dims.index("distance")
        new_sampling = step / 0.1
        new_coord = np.arange(start, stop, new_sampling)
        out = random_patch.interpolate(distance=new_coord)
        assert out.data.shape[axis] == len(new_coord)
        assert np.allclose(out.coords.get_array("distance"), new_coord)

    def test_uneven_sampling_rates(self, random_patch):
        """Uneven sampling should now work fine."""
        start, stop, step = get_start_stop_step(random_patch, "distance")
        new_samps = np.array([start, start + step, start + 4 * step, stop])
        out = random_patch.interpolate(distance=new_samps)
        assert not out.coords.coord_map["distance"].evenly_sampled

    def test_upsample_time(self, random_patch):
        """Ensure time can be upsampled."""
        start, stop, step = get_start_stop_step(random_patch, "time")
        axis = random_patch.dims.index("time")
        new = np.arange(start, stop, step / 2)
        out = random_patch.interpolate(time=new)
        assert out.attrs["time_min"] == np.min(new)
        assert out.attrs["time_max"] == np.max(new)
        assert out.attrs["time_step"] == np.mean(np.diff(new))
        assert out.data.shape[axis] == len(new)

    def test_endtime_updated(self, random_patch):
        """Ensure the endtime/starttime in coords and dims are consistent."""
        dist = [0, 42, 84, 126, 168, 210, 252, 294]
        out = random_patch.interpolate(distance=dist)
        coord = out.coords.get_array("distance")
        assert out.attrs["distance_max"] == coord.max()
        assert out.attrs["distance_min"] == coord.min()
        assert out.attrs["distance_step"] == np.median(np.diff(coord))

    def test_snap_like(self, wacky_dim_patch):
        """Ensure interpolate can be used to snapping coords."""
        patch = wacky_dim_patch.interpolate(time=None)
        time = patch.coords.coord_map["time"]
        assert time.evenly_sampled and time.sorted


class TestDecimate:
    """Ensure Patch can be decimated."""

    def test_decimate_no_lowpass(self, random_patch):
        """Simple decimation."""
        p1 = random_patch
        old_time = p1.coords.get_array("time")
        old_dt = old_time[1:] - old_time[:-1]
        # apply decimation,
        pa2 = random_patch.decimate(time=2)
        new_time = pa2.coords.get_array("time")
        new_dt = new_time[1:] - new_time[:-1]
        # ensure distance between time samples and shapes have changed
        len_ratio = np.round(len(old_dt) / len(new_dt))
        assert np.isclose(len_ratio, 2.0)
        dt_ratio = np.round(new_dt[0] / old_dt[0])
        assert np.isclose(dt_ratio, 2.0)

    def test_update_time_max(self, random_patch):
        """Ensure the time_max is updated after decimation."""
        out = random_patch.decimate(time=10)
        assert out.attrs["time_max"] == out.coords.get_array("time").max()

    def test_update_delta_dim(self, random_patch):
        """Since decimate changes the spacing of dimension this should be updated."""
        dt1 = random_patch.attrs.time_step
        out = random_patch.decimate(time=10)
        assert out.attrs["time_step"] == dt1 * 10

    def test_float_32_stability(self, random_patch):
        """
        Ensure float32 works for decimation.

        See scipy#15072.
        """
        ar = random_state.random((10_000, 2)).astype("float32")
        dt = dc.to_timedelta64(0.001)
        t1 = dc.to_datetime64("2020-01-01")
        coords = {
            "distance": np.array([1, 2]),
            "time": np.arange(0, ar.shape[0]) * dt + t1,
        }
        dims = ("time", "distance")
        attrs = {"time_step": dt, "time_min": t1}
        patch = dc.Patch(data=ar, coords=coords, dims=dims, attrs=attrs)
        # ensure all modes of decimation don't produce NaN values.
        decimated_iir = patch.decimate(time=10, filter_type="iir")
        assert not np.any(pd.isnull(decimated_iir.data))

        decimated_fir = patch.decimate(time=10, filter_type="fir")
        assert not np.any(pd.isnull(decimated_fir.data))

        decimated_none = patch.decimate(time=10, filter_type=None)
        assert not np.any(pd.isnull(decimated_none.data))

    def test_decimate_small_dimension(self, random_patch):
        """Ensure decimation raises helpful error on small dimensions."""
        small_patch = random_patch.select(distance=(0, 10), samples=True)
        match = "Scipy decimation failed."
        with pytest.raises(FilterValueError, match=match):
            small_patch.decimate(distance=2)


class TestResample:
    """Tests for resampling along a given dimension."""

    def test_downsample_time(self, random_patch):
        """Test decreasing the temporal sampling rate."""
        start, stop, step = get_start_stop_step(random_patch, "time")
        patch = random_patch
        axis = patch.dims.index("time")
        new_dt = 2 * step
        new = patch.resample(time=new_dt)
        assert new_dt == new.attrs["time_step"]
        assert np.all(np.diff(new.coords.get_array("time")) == new_dt)
        # ensure only the time dimension has changed.
        shape1, shape2 = random_patch.data.shape, new.data.shape
        for ax, (len1, len2) in enumerate(zip(shape1, shape2)):
            if ax == axis:  # Only resampled axis should have changed len
                assert len1 > len2
            else:
                assert len1 == len2

    def test_upsample_time(self, random_patch):
        """Test increasing the temporal sampling rate."""
        current_dt = random_patch.attrs["time_step"]
        axis = random_patch.dims.index("time")
        new_dt = current_dt / 2
        new = random_patch.resample(time=new_dt)
        assert new_dt == new.attrs["time_step"]
        assert np.all(np.diff(new.coords.get_array("time")) == new_dt)
        shape1, shape2 = random_patch.data.shape, new.data.shape
        for ax, (len1, len2) in enumerate(zip(shape1, shape2)):
            if ax == axis:  # Only resampled axis should have changed len
                assert len1 < len2
            else:
                assert len1 == len2

    def test_upsample_time_float(self, random_patch):
        """Test int as time sampling rate."""
        current_dt = random_patch.attrs["time_step"]
        axis = random_patch.dims.index("time")
        new_dt = current_dt / 2
        new = random_patch.resample(time=new_dt / np.timedelta64(1, "s"))
        assert new_dt == new.attrs["time_step"]
        assert np.all(np.diff(new.coords.get_array("time")) == new_dt)
        shape1, shape2 = random_patch.data.shape, new.data.shape
        for ax, (len1, len2) in enumerate(zip(shape1, shape2)):
            if ax == axis:  # Only resampled axis should have changed len
                assert len1 < len2
            else:
                assert len1 == len2

    def test_resample_distance(self, random_patch):
        """Ensure distance dimension is also resample-able."""
        current_dx = random_patch.attrs["distance_step"]
        new_dx = current_dx / 2
        new = random_patch.resample(distance=new_dx)
        axis = random_patch.dims.index("distance")
        assert new_dx == new.attrs["distance_step"]
        assert np.allclose(np.diff(new.coords.get_array("distance")), new_dx)
        shape1, shape2 = random_patch.data.shape, new.data.shape
        for ax, (len1, len2) in enumerate(zip(shape1, shape2)):
            if ax == axis:  # Only resampled axis should have changed len
                assert len1 < len2
            else:
                assert len1 == len2

    def test_odd_sampling_rate(self, random_patch):
        """Tests for resampling to a non-int sampling rate."""
        new_step = 1.232132323222
        out = random_patch.resample(distance=new_step)
        assert out.attrs["distance_max"] <= random_patch.attrs["distance_max"]
        assert np.allclose(out.attrs["distance_step"], new_step)

    def test_slightly_above_current_rate(self, random_patch):
        """Tests for resampling slightly above current rate."""
        start, stop, step = get_start_stop_step(random_patch, "distance")
        new_step = step + 0.0000001
        out = random_patch.resample(distance=new_step)
        assert out.attrs["distance_max"] <= random_patch.attrs["distance_max"]
        assert np.allclose(out.attrs["distance_step"], new_step)

    def test_slightly_under_current_rate(self, random_patch):
        """Tests for resampling slightly under current rate."""
        start, stop, step = get_start_stop_step(random_patch, "distance")
        new_step = step - 0.0000001
        out = random_patch.resample(distance=new_step)
        assert out.attrs["distance_max"] <= random_patch.attrs["distance_max"]
        assert np.allclose(out.attrs["distance_step"], new_step)

    def test_odd_time(self, random_patch):
        """Tests resampling to odd time interval."""
        dt = np.timedelta64(1234567, "ns")
        out = random_patch.resample(time=dt)
        new_dt = out.attrs["time_step"]
        assert np.isclose(float(new_dt), float(dt))
        assert out.attrs

    def test_huge_resample(self, random_patch):
        """Tests for greatly increasing the sampling_period."""
        out = random_patch.resample(distance=42, samples=True)
        assert len(out.coords.get_array("distance")) == 42

    def test_resample_with_units_hz(self, random_patch):
        """Ensure resample works with units."""
        new1 = random_patch.resample(time=50 * Hz)
        new2 = random_patch.resample(time=1 / 50)
        new3 = random_patch.resample(time=1 / 50 * s)
        assert new1 == new2 == new3

    def test_resample_distance_with_units(self, random_patch):
        """Ensure distance can be resampled with units as well."""
        new1 = random_patch.resample(distance=5 * m)
        new2 = random_patch.resample(distance=1 / 5 * (1 / m))
        new3 = random_patch.resample(distance=5)
        assert new1 == new2 == new3

    def test_resample_docs(self, random_patch):
        """Ensure docstring examples runs."""
        patch = random_patch
        time = patch.coords.get_array("time")
        ts = patch.attrs.time_step
        new_time = np.arange(time.min(), time.max(), 0.5 * ts)
        uptime = patch.interpolate(time=new_time)
        assert isinstance(uptime, dc.Patch)
        # interpolate unevenly sampled dim to evenly sampled
        patch = dc.get_example_patch("wacky_dim_coords_patch")
        patch_time_even = patch.interpolate(time=None)
        assert isinstance(patch_time_even, dc.Patch)

    def test_iresample_time(self, random_patch):
        """Tests iresample in time dim."""
        time_samples = 40
        out = random_patch.resample(time=time_samples, samples=True)
        assert len(out.coords.get_array("time")) == time_samples

    def test_iresample_distance(self, random_patch):
        """Test for resampling distance to set len."""
        dist = 42
        out = random_patch.resample(distance=dist, samples=True)
        assert len(out.coords.get_array("distance")) == dist

    def test_iresample_deprecated(self, random_patch):
        """Ensure iresample issues deprecation warning."""
        with pytest.warns(DeprecationWarning):
            random_patch.iresample(distance=42)

    def test_resample_fft(self, random_patch):
        """Tests for resample rft axis. See #272."""
        out = random_patch.dft("time", real="time").resample(ft_time=1)
        assert isinstance(out, dc.Patch)



"""
Utilities for reading binary sintela data.

Notes
-----
- We make the strong assumption that each file contains contiguous
  blocks of DAS data.

- We only support v3 for now, though some commented out code here has the
  start for version 4 support.
"""

import numpy as np

import dascore as dc
from dascore.compat import array
from dascore.core.attrs import PatchAttrs
from dascore.core.coordmanager import get_coord_manager
from dascore.core.coords import get_coord
from dascore.exceptions import InvalidFiberFileError
from dascore.utils.misc import get_buffer_size, maybe_mem_map

SYNC_WORD = 0x11223344
MAX_TRIGGERS = 8
DIMS = ("time", "distance")

# The minimum required info to determine if a file is a Sintela binary file.
base_header_dtypes = np.dtype(
    [
        ("sync_word", "<u4"),
        ("header_size", "<u4"),
        ("version", "<u4"),
    ]
)

# numpy dtypes version 3 of header (excluding base)
_HEADER_DTYPES = {
    "3": np.dtype(
        [
            ("sample_count", "<u8"),
            ("start_time_ns", "<u8"),
            ("num_channels", "<u4"),
            ("num_samples", "<u4"),
            ("sample_rate", "<f4"),
            ("channel_spacing", "<f4"),
            ("gauge_length", "<f4"),
            ("trigger", "<u4"),
            ("start_channel", "<u4"),
            ("channel_step", "<u4"),
            ("gps_pps_time", "<u4"),
            ("gps_pps_sample_offset", "<i4"),
            ("gps_status", "<u4"),
            ("trigger_flags", "<u1"),
            ("status_flags", "<u1"),
            ("axial_counter_states", "<u1"),
            ("trigger_type", "<u1"),
            ("data_type", "<u1"),
            ("demod_data_type", "<u1"),
            ("unused", "<u2"),
            ("trigger_offsets", "<u2", (MAX_TRIGGERS,)),
        ]
    ),
    # "4": np.dtype([
    #     ("codec_type", "|S4"),
    #     ("sample_count", "<u8"),
    #     ("start_time_ns", "<u8"),
    #     ("num_channels", "<u4"),
    #     ("num_samples", "<u4"),
    #     ("sample_rate", "<f4"),
    #     ("channel_spacing", "<f4"),
    #     ("gauge_length", "<f4"),
    #     ("trigger", "<u4"),
    #     ("start_channel", "<u4"),
    #     ("channel_step", "<u4"),
    #     ("scale_factor", "<f4"),
    #     ("align", "<i4"),
    #     ("gps_pps_time", "<i4"),
    #     ("gps_pps_offset", "<i4"),
    #     ("gps_status", "<i4"),
    #     ("flags", "u1", (8,)),
    #     ("trigger_offsets", "<u2", (MAX_TRIGGERS,)),
    #     ("spare", "<i4"),
    # ]),
}

# The size of the headers
_HEADER_SIZES = {
    "3": base_header_dtypes.itemsize + _HEADER_DTYPES["3"].itemsize,
    # '4': base_header_dtypes.itemsize + _HEADER_DTYPES['4'].itemsize,
}

# The types of data that can be contained in file and their units.
_DATA_TYPE_MAP = {
    0: ("phase", "radians"),
    1: ("phase_difference", "radians"),
    2: ("phase_rate", "radians/s"),
    3: ("strain", "microstrain"),
    4: ("strain_rate", "microstrain/s"),
}


def _read_base_header(fid):
    """Return the first 3 elements of the sintela header."""
    array = np.fromfile(fid, dtype=base_header_dtypes, count=1)
    out = {x: y for x, y in zip(array.dtype.names, array[0])}
    return out


def _get_number_of_packets(fid, header, size):
    """Get the number of packets in the file."""
    # get filesize without depending on input being a file.
    file_size = get_buffer_size(fid)
    samples = header["num_samples"]
    channels = header["num_channels"]
    packet_size = size + (channels * samples * 4)
    num_packets, remainder = divmod(file_size, packet_size)
    if remainder:
        msg = "Sintela binary file size not divisible by packet size."
        raise InvalidFiberFileError(msg)
    return num_packets


def _read_remaining_header(fid, base):
    """Read file header using numpy structured dtype."""
    version = str(base["version"])
    header_size = base["header_size"]
    expected_size = _HEADER_SIZES.get(version, -1)
    assert header_size == expected_size
    dtype = _HEADER_DTYPES[version]
    buf = np.fromfile(fid, dtype=dtype, count=1)
    header = {x: y for x, y in zip(buf.dtype.names, buf[0])}
    assert version == "3", "only 3 support for now,"
    header["num_packets"] = _get_number_of_packets(fid, header, header_size)
    header["dtype"] = "<f4"
    return header


def _get_complete_header(rid):
    """Get the complete header information."""
    header = _read_base_header(rid)
    header.update(_read_remaining_header(rid, header))
    return header


def _get_time_coord(header):
    """Get the time coordinate."""
    starttime = np.asarray(header["start_time_ns"]).astype("datetime64[ns]")[()]
    timestep = dc.to_timedelta64(1 / header["sample_rate"])
    total_len = header["num_packets"] * header["num_samples"]
    end_time = starttime + timestep * total_len
    time = get_coord(start=starttime, step=timestep, stop=end_time)
    assert len(time) == total_len
    return time


def _get_dist_coord(header):
    """Get the distance coordinate."""
    channel_step = header["channel_step"]
    channel_count = header["num_channels"]
    start_channel = header["start_channel"]
    dx = header["channel_spacing"]
    dist_start = start_channel * dx
    coord = get_coord(
        start=dist_start,
        step=channel_step * dx,
        stop=dist_start + (channel_count * channel_step) * dx,
        units="m",
    )
    assert len(coord) == channel_count
    return coord


def _get_attr_dict(header, extras=None):
    """
    Extract info from sintela headers which will be used to create patch Attrs.
    """
    data_type, units = _DATA_TYPE_MAP.get(header["data_type"], ("phase", "radians"))
    out = dict(
        data_type=data_type,
        data_units=units,
        gauge_length=header["gauge_length"],
        gps_status=header["gps_status"],
    )
    out.update(extras if extras is not None else {})
    return out


def _load_data(fid, header):
    """Use numpy's memmap to get array information."""
    header_size = header["header_size"]
    num_channels = header["num_channels"]
    num_samples = header["num_samples"]
    dtype = np.dtype(header["dtype"])
    packet_size = header_size + num_channels * num_samples * dtype.itemsize
    # Memory map entire file as raw bytes. We do this to later skip the headers
    fid.seek(0)
    raw = maybe_mem_map(fid)
    # Compute how many blocks
    block_count = raw.size // packet_size
    # Create a view that skips headers
    data = raw.reshape(block_count, packet_size)[:, header_size:]
    data = data.view(dtype).reshape(-1, num_channels)
    return data


def _get_attrs_coords_header(rid, attr_class=PatchAttrs, extras=None):
    """Get Patch attributes and coordinates."""
    header = _get_complete_header(rid)
    coords = {"time": _get_time_coord(header), "distance": _get_dist_coord(header)}
    cm = get_coord_manager(coords=coords, dims=DIMS)
    attrs = _get_attr_dict(header, extras)
    attrs["coords"] = cm
    return attr_class(**attrs), cm, header


def _get_patch(
    resource, attr_class=PatchAttrs, extras=None, time=None, distance=None, **kwargs
):
    """Get the patch from the sintela file."""
    patch_attrs, cm, header = _get_attrs_coords_header(resource, attr_class, extras)
    data = _load_data(resource, header)
    # Apply slicing if needed. This is done before patch creation so a memmap
    # can be sliced before loading all data into memory.
    if time is not None or distance is not None:
        cm, data = cm.select(data, time=time, distance=distance)
    patch = dc.Patch(data=array(data), coords=cm, attrs=patch_attrs)
    return patch



"""Module for calculating cross-correlation over time or distance."""

from __future__ import annotations

import warnings

import numpy as np

import dascore as dc
from dascore.constants import PatchType
from dascore.utils.patch import (
    get_dim_axis_value,
    patch_function,
)
from dascore.utils.time import to_float


def _get_source_fft(patch, dim, source, source_axis, samples):
    """
    Get an array of coordinate sources.

    This function will place the new sources in a third dimension so
    they broadcast with the original fft matrix.
    """
    # Extract an array containing just the sources
    coord_source = patch.get_coord(dim)
    index_source = coord_source.get_next_index(source, samples=samples)
    selecter = [slice(None), slice(None), None]
    selecter[source_axis] = np.atleast_1d(index_source)
    source = patch.data[tuple(selecter)]
    # Now transpose source so source dim is list. Essentially we just
    # need to swap the source axis with the last axis.
    out = np.swapaxes(source, source_axis, -1)
    return out


@patch_function()
def correlate_shift(patch, dim, undo_weighting=True):
    """
    Apply a shift to the patch data to undo correlation in frequency domain.

    Also adds the appropriate coordinate prefixed with "lag" and has a datatype
    of float.

    Parameters
    ----------
    patch
        The input patch
    dim
        The dimension name that was correlated in the freq. domain.
    undo_weighting
        If True, also undo the weighting artifact caused by DASCore's dft
        weighting. This is done by simply dividing by the coordinate step.
        See [dft note](`docs/notes/dft_notes.qmd`) for more details.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>>
    >>> # Example 1
    >>> # An auto-correlation of the example patch
    >>> dft = patch.dft("time", real=True)
    >>> dft_sq = dft * dft.conj()
    >>> idft = dft_sq.idft()
    >>> auto_patch = idft.correlate_shift(dim="time")
    """
    coord = patch.get_coord(dim, require_evenly_sampled=True)
    axis = patch.dims.index(dim)
    data = np.fft.fftshift(patch.data, axes=axis)
    if undo_weighting:
        data = data / to_float(coord.step)
    step = coord.step
    new_start = -np.ceil((len(coord) - 1) / 2) * step
    new_end = np.ceil((len(coord) - 1) / 2) * step
    _new_coord = dc.get_coord(
        start=new_start, stop=new_end, step=step, units=coord.units
    )
    new_coord = _new_coord.change_length(len(coord))
    assert len(new_coord) == len(coord)
    cm = patch.coords
    new_cm = cm.update(**{dim: new_coord}).rename_coord(**{dim: f"lag_{dim}"})
    out = patch.update(data=data, coords=new_cm)
    return out


@patch_function()
def correlate(
    patch: PatchType,
    samples=False,
    lag=None,
    **kwargs,
) -> PatchType:
    """
    Correlate source row/columns in a 2D patch with all other row/columns.

    Correlations are done in the frequency domain. This function can accept a
    patch whose target dimension has already been transformed with the
    [`Patch.dft`](`dascore.transform.fourier.dft`) method, otherwise the dft
    will be performed. If the input has already been transformed,
    [`Patch.correlation_shift`](`dascore.proc.correlate.correlate_shift`)
    is useful to undo dft artefacts after the idft is applied.

    While a 2D patch is required for input, a 3D patch is returned where the
    3rd dimesion cooresponds to the source rows/columns. For the case of a
    single source, the [`Patch.squeeze`](`dascore.Patch.squeeze`) method
    can be helpful to remove length 1 dimensions.

    Parameters
    ----------
    patch : PatchType
        The input data patch to be cross-correlated. Must be 2-dimensional.
        The patch can be in time or frequency domains.
    samples : bool, optional (default = False)
        If True, the argument specified in kwargs refers to the *sample* not
        value along that axis. See examples for details.
    lag
        Deprecated, just use select on the output patch instead.
    **kwargs
        Specifies correlation dimension and the master source(s), to which
        we want to cross-correlate all other channels/time samples.If the
        master source is an array, the function will compute correlations for
        all the posible pairs.

    Examples
    --------
    >>> import dascore as dc
    >>> from dascore.units import m, s

    >>> # Get a patch composed of sin waves whose correlation results
    >>> # can easily be checked.
    >>> patch = dc.get_example_patch(
    ...     "sin_wav",
    ...     sample_rate=100,
    ...     frequency=range(10, 20),
    ...     duration=5,
    ...     channel_count=10,
    ... ).taper(time=0.5).set_units(distance='m')
    >>>
    >>> # Example 1
    >>> # Calculate cc for all channels as receivers and
    >>> # the 10 m channel as the master channel. Squeeze the output
    >>> # so the returned patch is 2D.
    >>> cc_patch = patch.correlate(distance = 10 * m).squeeze()
    >>>
    >>> # Example 2
    >>> # Get cc within (-2,2) sec of lag for all channels as receivers
    >>> # and the 10 m channel as the master channel. The new patch has dimensions
    >>> # (lag_time, distance, source_distance)
    >>> cc_patch = (
    ...     patch.correlate(distance = 10 * m)
    ...     .select(lag_time=(-2, 2))
    ... )
    >>>
    >>> # Example 3
    >>> # First remove every other distance channel (less memory usage)
    >>> # the use the new 2nd channel as the source.
    >>> cc_patch = (
    ...     patch.decimate(distance=2, filter_type=None)
    ...     .correlate(distance=1, samples=True)
    ... )
    >>>
    >>> # Example 4
    >>> # Correlate along time dimension (perhaps for template matching
    >>> # applications)
    >>> cc_patch = patch.correlate(time=100, samples=True)
    >>>
    >>> # Example 5
    >>> # A pipeline of frequency domain correlation and an array of sources
    >>> padded_patch = patch.pad(time="correlate")  # pad to at least 2n + 1
    >>> dft_patch = patch.dft("time", real=True)
    >>> # Any other pre-processing steps go here...
    >>> # ...
    >>> # Perform the correlation with 3 source channels
    >>> cc_patch = dft_patch.correlate(distance=[1, 3, 7], samples=True)
    >>> # Perform any post-processing here
    >>> # ...
    >>> # Convert back to time domain, apply `correlate shift` to undo
    >>> # fft related shifting and scaling as well as create lag coordinate.
    >>> cc_out = cc_patch.idft().correlate_shift("time")

    Notes
    -----
    1 - The cross-correlation is performed in the frequency domain.

    2 - The output dimension is opposite of the one specified in kwargs and
      shares a name with the original coord except the string "lag_" is
      prepended. For example, "lag_time".
    """
    if lag is not None:
        msg = "Correlate lag is deprecated. Simply use on the output patch."
        warnings.warn(msg, DeprecationWarning)
    assert len(patch.dims) == 2, "must be a 2D patch."
    dim, source_axis, source = get_dim_axis_value(patch, kwargs=kwargs)[0]
    # Get the axis and coord over which fft should be calculated.
    fft_axis = next(iter(set(range(len(patch.dims))) - {source_axis}))
    fft_dim = patch.dims[fft_axis]
    # Determine if the input patch has already been transformed.
    input_dft = fft_dim.startswith("ft_")
    is_real = not np.issubdtype(patch.data.dtype, np.complexfloating)
    if not input_dft:  # Standard dft workflow for correlation
        # Note: we use .func here to avoid getting these added to the history.
        padded = patch.pad.func(patch, **{fft_dim: "correlate"})
        patch = padded.dft.func(padded, fft_dim, real=fft_dim if is_real else None)
    # Get the sources.
    source = patch.get_coord(dim).values if source is None else source
    source_fft = _get_source_fft(patch, dim, source, source_axis, samples)
    # Need to insert new axis so the arrays broadcast correctly.
    fft_patch_array = patch.data[..., None]
    fft_prod = fft_patch_array * np.conj(source_fft)
    # Create frequency domain patch with results
    source = getattr(source, "magnitude", source)  # strips units
    new_coord = dc.get_coord(values=np.atleast_1d(source))
    dim_name = f"source_{dim}"
    cm = patch.coords.update(**{dim_name: (dim_name, new_coord)})
    out = patch.update(data=fft_prod, coords=cm)
    # Undo fft if this function did one, shift, and update coord.
    if not input_dft:
        idft = out.idft.func(out)
        out = idft.correlate_shift.func(idft, fft_dim)
    return out



"""
RSF format support module.

Notes
-----
- output has been tested on Madagascar version 3.1-git


Examples
--------
import dascore as dc

# get the path to a random DAS file.
patch = dc.get_example_patch()
patch.io.write("test_out.rsf","rsf",data_path="test_out.rsf")

"""

from .core import RSFV1



"""
Tests for optoDAS files.
"""

import dascore as dc
from dascore.io.optodas import OptoDASV8
from dascore.utils.downloader import fetch


class TestOptoDASIssues:
    """Test case related to issues in OptoDAS parser."""

    def test_read_decimated_patch(self):
        """Tests for reading spatially decimated patch (#419)"""
        path = fetch("decimated_optodas.hdf5")
        fiber_io = OptoDASV8()

        fmt_str, version_str = fiber_io.get_format(path)
        assert (fmt_str, version_str) == (fiber_io.name, fiber_io.version)

        spool = fiber_io.read(path)
        patch = spool[0]
        assert isinstance(patch, dc.Patch)
        assert patch.data.shape



"""Misc. tests for misfit utilities."""

from __future__ import annotations

import os
import time
import warnings
from io import BytesIO
from pathlib import Path

import numpy as np
import pytest

from dascore.exceptions import MissingOptionalDependencyError
from dascore.utils.misc import (
    MethodNameSpace,
    _iter_filesystem,
    cached_method,
    get_buffer_size,
    get_stencil_coefs,
    iterate,
    maybe_get_items,
    maybe_mem_map,
    optional_import,
    to_object_array,
    warn_or_raise,
)


class ParentClass:
    """A test parent class."""

    @property
    def namespace(self):
        """Your run-o-the-mill namespace."""
        return MNS(self)


class MNS(MethodNameSpace):
    """method name space subclass."""

    def func1(self, expected_type):
        """First func."""
        return isinstance(self, expected_type)


class TestNamespaceClass:
    """Tests for namespace class."""

    def test_parent_self_passed_to_namespace(self):
        """Ensure the parent of namespace gets passed to self."""
        pc = ParentClass()
        assert pc.namespace.func1(ParentClass)

    def test_assign_adhoc_method(self):
        """Ensure methods added after class definition still work."""

        def new_method(self, expected_type):
            return isinstance(self, expected_type)

        MNS.new_method = new_method
        pc = ParentClass()
        assert pc.namespace.new_method(ParentClass)


class TestIterFS:
    """Tests for iterating directories of files."""

    sub = {"D": {"C": ".mseed"}, "F": ".json", "G": {"H": ".txt"}}  # noqa
    file_paths = {"A": ".txt", "B": sub}  # noqa

    # --- helper functions
    def setup_test_directory(self, some_dict: dict, path: Path):
        """Build the test directory."""
        for path in self.get_file_paths(some_dict, path):
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("w") as fi:
                fi.write("useful text")

    def get_file_paths(self, some_dict, path):
        """Return expected paths to files."""
        for i, v in some_dict.items():
            if isinstance(v, dict):
                yield from self.get_file_paths(v, path / i)
            else:
                yield path / (i + v)

    # --- fixtures
    @pytest.fixture(scope="class")
    def simple_dir(self, tmp_path_factory):
        """Return a simple directory for iterating."""
        path = Path(tmp_path_factory.mktemp("iterfiles"))
        self.setup_test_directory(self.file_paths, path)
        return path

    @pytest.fixture(scope="class")
    def dir_with_hidden_dir(self, tmp_path_factory):
        """Create a directory with a hidden directory inside."""
        path = Path(tmp_path_factory.mktemp("iterfiles_hidden"))
        struct = dict(self.file_paths)
        # add hidden directory with files in it.
        struct[".Hidden"] = {"Another": {"hidden_by_parent": ".txt"}}
        self.setup_test_directory(struct, path)
        return path

    def test_basic(self, simple_dir):
        """Test basic usage of iterfiles."""
        files = set(self.get_file_paths(self.file_paths, simple_dir))
        out = {Path(x) for x in _iter_filesystem(simple_dir)}
        assert files == out

    def test_one_subdir(self, simple_dir):
        """Test with one sub directory."""
        subdirs = simple_dir / "B" / "D"
        out = set(_iter_filesystem(subdirs))
        assert len(out) == 1

    def test_multiple_subdirs(self, simple_dir):
        """Test with multiple sub directories."""
        path1 = simple_dir / "B" / "D"
        path2 = simple_dir / "B" / "G"
        out = {Path(x) for x in _iter_filesystem([path1, path2])}
        files = self.get_file_paths(self.file_paths, simple_dir)
        expected = {
            x
            for x in files
            if str(x).startswith(str(path1)) or str(x).startswith(str(path2))
        }
        assert out == expected

    def test_extension(self, simple_dir):
        """Test filtering based on extention."""
        out = set(_iter_filesystem(simple_dir, ext=".txt"))
        for val in out:
            assert val.endswith(".txt")

    def test_mtime(self, simple_dir):
        """Test filtering based on modified time."""
        files = list(self.get_file_paths(self.file_paths, simple_dir))
        # set the first file mtime in future
        now = time.time()
        first_file = files[0]
        os.utime(first_file, (now + 10, now + 10))
        # get output make sure it only returned first file
        out = list(_iter_filesystem(simple_dir, timestamp=now + 5))
        assert len(out) == 1
        assert Path(out[0]) == first_file

    def test_skips_files_in_hidden_directory(self, dir_with_hidden_dir):
        """Hidden directory files should be skipped."""
        out1 = list(_iter_filesystem(dir_with_hidden_dir))
        has_hidden_by_parent = ["hidden_by_parent" in x for x in out1]
        assert not any(has_hidden_by_parent)
        # But if skip_hidden is False it should be there
        out2 = list(_iter_filesystem(dir_with_hidden_dir, skip_hidden=False))
        has_hidden_by_parent = ["hidden_by_parent" in x for x in out2]
        assert sum(has_hidden_by_parent) == 1

    def test_pass_file(self, dummy_text_file):
        """Just pass a single file and ensure it gets returned."""
        out = list(_iter_filesystem(dummy_text_file))
        assert len(out) == 1
        assert out[0] == dummy_text_file

    def test_no_directories(self, simple_dir):
        """Ensure no directories are included when include_directories=False."""
        out = list(_iter_filesystem(simple_dir, include_directories=False))
        has_dirs = [Path(x).is_dir() for x in out]
        assert not any(has_dirs)

    def test_include_directories(self, simple_dir):
        """Ensure we can get directories back."""
        out = list(_iter_filesystem(simple_dir, include_directories=True))
        returned_dirs = [Path(x) for x in out if Path(x).is_dir()]
        assert len(returned_dirs)
        # The top level directory should have been included
        assert simple_dir in returned_dirs
        # Directory names
        dir_names = {x.name for x in returned_dirs}
        expected_names = {"B", "G", "D"}
        assert expected_names.issubset(dir_names)

    def test_skip_signal_directory(self, simple_dir):
        """Ensure a skip signal can be sent to stop parsing on directory."""
        out = []
        iterator = _iter_filesystem(simple_dir, include_directories=True)
        for path in iterator:
            if Path(path).name == "B":
                iterator.send("skip")
            out.append(path)
        names = {Path(x).name.split(".")[0] for x in out}
        # Anything after B should have been skipped
        assert {"C", "D", "E", "F"}.isdisjoint(names)


class TestIterate:
    """Test case for iterate."""

    def test_none(self):
        """None should return an empty tuple."""
        assert iterate(None) == tuple()

    def test_object(self):
        """A single object should be returned in a tuple."""
        assert iterate(1) == (1,)

    def test_str(self):
        """A single string object should be returned as a tuple."""
        assert iterate("hey") == ("hey",)


class TestOptionalImport:
    """Ensure the optional import works."""

    def test_import_installed_module(self):
        """Test to ensure an installed module imports."""
        import dascore as dc

        mod = optional_import("dascore")
        assert mod is dc
        sub_mod = optional_import("dascore.core")
        assert sub_mod is dc.core

    def test_missing_module_raises(self):
        """Ensure a module which is missing raises the appropriate Error."""
        with pytest.raises(MissingOptionalDependencyError, match="boblib4"):
            optional_import("boblib4")


class TestGetStencilCoefficients:
    """Tests for stencil coefficients."""

    def test_3_point_1st_derivative(self):
        """3 point 1st derivative."""
        out = get_stencil_coefs(1, 1)
        expected = np.array([-1 / 2, 0, 1 / 2])
        assert np.allclose(out, expected)

    def test_5_point_1st_derivative(self):
        """5 point 1st derivative."""
        out = get_stencil_coefs(2, 1)
        expected = np.array([1, -8, 0, 8, -1]) / 12.0
        assert np.allclose(out, expected)

    def test_3_point_2nd_derivative(self):
        """3 point 2nd derivative."""
        out = get_stencil_coefs(1, 2)
        expected = np.array([1, -2, 1])
        assert np.allclose(out, expected)


class TestCachedMethod:
    """Ensure cached methods caches method calls (duh)."""

    class _JohnnyCached:
        """Tests class for caching."""

        @cached_method
        def no_args(self):
            """No argument method."""
            return {"output", "defined"}

        @cached_method
        def multiargs(self, a, b):
            """Multiple arguments for cache testing."""
            return a + b

    def test_no_args_kwargs(self):
        """Ensure objects cache without args or kwargs."""
        john = self._JohnnyCached()
        first = john.no_args()
        assert first is john.no_args()

    def test_positional(self):
        """Ensure positional arguments work."""
        john = self._JohnnyCached()
        out = john.multiargs(1, 2)
        assert john.multiargs(1.0, 2.0) == 3.0
        assert out == 3

    def test_kwargs(self):
        """Ensure keywords also work."""
        john = self._JohnnyCached()
        assert john.multiargs(1, b=1) == 2
        assert john.multiargs(a=2, b=3) == 5


class TestMaybeGetItems:
    """Tests for maybe_get_attrs."""

    def test_missed_itme(self):
        """Ensure it still works when a key is missing."""
        data = {"bob": 1, "bill": 2}
        expected = {"bob": "sue", "lary": "who"}
        out = maybe_get_items(data, attr_map=expected)
        assert "sue" in out
        assert "who" not in out


class TestWarnOrRaise:
    """Ensure warn or raise works."""

    def test_warn(self):
        """Ensure a warning is emitted."""
        msg = "Warning: this is a warning"
        with pytest.warns(UserWarning, match=msg):
            warn_or_raise(msg, warning=UserWarning, behavior="warn")

    def test_raise(self):
        """Ensure an exception can be raised."""
        msg = "Something went wrong."
        with pytest.raises(ValueError, match=msg):
            warn_or_raise(msg, exception=ValueError, behavior="raise")

    def test_nothing(self):
        """Ensure when  None does nothing."""
        msg = "Big nothing burger"
        # Now exceptions or warnings will crash the program.
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            warn_or_raise(msg, behavior=None)


class TestToObjectArray:
    """Tests for converting a sequence of objects to an object array."""

    def test_patches_to_array(self, random_patch):
        """Ensure a list of patches can be converted to an object array."""
        patches = [random_patch] * 3
        out = to_object_array(patches)
        assert isinstance(out, np.ndarray)


class TestGetBufferSize:
    """Ensure we can get the size of various buffers."""

    def test_file(self, terra15_v5_path):
        """Ensure we can get the size of various buffers."""
        expected_size = Path(terra15_v5_path).stat().st_size
        with open(terra15_v5_path, "rb") as fid:
            out = get_buffer_size(fid)
        assert out == expected_size

    def test_bytes_io(self):
        """Ensure it can also get size of bytes io."""
        bio = BytesIO()
        bio.write(b"1234")
        size1 = get_buffer_size(bio)
        bio.seek(0)
        size2 = get_buffer_size(bio)
        assert size1 == size2 == 4


class TestMaybeMemMap:
    """Ensure we can get byte arrays from various objects."""

    def test_file(self, terra15_v5_path):
        """Test files return memmap."""
        with open(terra15_v5_path, "rb") as fid:
            array = maybe_mem_map(fid)
        assert isinstance(array, np.memmap)
        assert array.size

    def test_bytes_io(self):
        """Ensure non-files return arrays."""
        bio = BytesIO()
        bio.write(b"1234")
        bio.seek(0)
        array = maybe_mem_map(bio)
        assert isinstance(array, np.ndarray)
        assert array.size == 4

    def test_bytes_io_nonzero_position(self):
        """Fallback should read entire buffer even if pointer is not at 0."""
        bio = BytesIO()
        bio.write(b"abcdef")
        # Intentionally do not seek back to 0; pointer is at end.
        arr_end_pos = maybe_mem_map(bio)
        # Now reset to start and compare lengths; both should see full content.
        bio.seek(0)
        arr_full = maybe_mem_map(bio)
        assert arr_end_pos.size == arr_full.size == 6



"""
Core module for reading Sintela binary format.
"""

from __future__ import annotations

import numpy as np

import dascore as dc
from dascore.constants import opt_timeable_types
from dascore.io import FiberIO
from dascore.utils.io import BinaryReader

from .utils import (
    _HEADER_SIZES,
    SYNC_WORD,
    _get_attrs_coords_header,
    _get_patch,
    _read_base_header,
)


class SintelaPatchAttrs(dc.PatchAttrs):
    """Patch Attributes for Sintela binary format."""

    gauge_length: float = np.nan
    gauge_length_units: str = "m"


class SintelaBinaryV3(FiberIO):
    """Version 3 of Sintela's binary format."""

    name = "Sintela_Binary"
    preferred_extensions = ("raw",)
    version = "3"

    def get_format(self, resource: BinaryReader, **kwargs) -> tuple[str, str] | bool:
        """
        Return name and version string or False.

        Parameters
        ----------
        resource
            A path to the file which may contain terra15 data.
        """
        resource.seek(0)
        base = _read_base_header(resource)
        sync = base["sync_word"]
        version = str(base["version"])
        size = base["header_size"]
        expected_size = _HEADER_SIZES.get(version, 0)
        if sync == SYNC_WORD and version == self.version and size == expected_size:
            return self.name, version
        return False

    def scan(self, resource: BinaryReader, **kwargs) -> list[dc.PatchAttrs]:
        """Scan a file, return summary information on the contents."""
        extras = {
            "path": resource.name,
            "file_format": self.name,
            "file_version": self.version,
        }
        attrs, _, _ = _get_attrs_coords_header(
            resource, SintelaPatchAttrs, extras=extras
        )

        return [attrs]

    def read(
        self,
        resource: BinaryReader,
        time: tuple[opt_timeable_types, opt_timeable_types] | None = None,
        distance: tuple[float | None, float | None] | None = None,
        **kwargs,
    ) -> dc.BaseSpool:
        """Read a single Sintela binary file."""
        patch = _get_patch(
            resource, time=time, distance=distance, attr_class=SintelaPatchAttrs
        )
        return dc.spool(patch)



"""Tests for the GDR file format."""

import numpy as np
import pytest

from dascore.io.gdr import GDR_V1
from dascore.utils.downloader import fetch


@pytest.fixture(scope="module")
def gpr_path():
    """Return the file path to a GDR file."""
    return fetch("gdr_1.h5")


class TestGDR:
    """Misc. tests not covered by common tests."""

    def test_no_snap(self, gpr_path):
        """Ensure snap or no snap produces the same coord for this file."""
        fiber_io = GDR_V1()
        patch1 = fiber_io.read(gpr_path, snap=False)[0]
        patch2 = fiber_io.read(gpr_path, snap=True)[0]
        time_1 = patch1.get_coord("time")
        time_2 = patch2.get_coord("time")
        assert len(time_1) == len(time_2)
        assert np.all(time_1.values == time_2.values)



"""Tests for coordinate object."""

from __future__ import annotations

import datetime
import pickle
from functools import partial
from io import BytesIO
from typing import ClassVar

import numpy as np
import pandas as pd
import pytest
import rich.text
from pydantic import ValidationError

import dascore as dc
from dascore.compat import random_state
from dascore.core.coords import (
    BaseCoord,
    CoordMonotonicArray,
    CoordPartial,
    CoordRange,
    CoordSummary,
    get_coord,
)
from dascore.exceptions import CoordError, ParameterError
from dascore.units import get_quantity
from dascore.utils.misc import all_close, register_func
from dascore.utils.time import dtype_time_like, is_datetime64, is_timedelta64, to_float

COORDS = []


@pytest.fixture(scope="session")
@register_func(COORDS)
def evenly_sampled_coord():
    """Create coordinates which are evenly sampled."""
    ar = np.arange(0, 100, 1)
    return get_coord(data=ar)


@pytest.fixture(scope="session")
@register_func(COORDS)
def evenly_sampled_reversed_coord():
    """Create coordinates which are evenly sampled."""
    ar = np.arange(0, -100, -1)
    return get_coord(data=ar)


@pytest.fixture(scope="session")
@register_func(COORDS)
def evenly_sampled_float_coord_with_units():
    """Create coordinates which are evenly sampled and have units."""
    ar = np.arange(1, 100, 1 / 3)
    return get_coord(data=ar, units="m")


@pytest.fixture(scope="session")
@register_func(COORDS)
def evenly_sampled_float_reverse_units_coord():
    """Create coordinates which are evenly sampled and have units."""
    ar = np.arange(-1, -100, -1 / 3)
    return get_coord(data=ar, units="m")


@pytest.fixture(scope="session")
@register_func(COORDS)
def evenly_sampled_date_coord():
    """Create coordinates which are evenly sampled."""
    ar = dc.to_datetime64(np.arange(1, 100_000, 1_000))
    return get_coord(data=ar)


@pytest.fixture(scope="session")
@register_func(COORDS)
def evenly_sampled_time_delta_coord():
    """Create coordinates which are evenly sampled."""
    ar = dc.to_timedelta64(np.arange(1, 100_000, 1_000))
    return get_coord(data=ar)


@pytest.fixture(scope="session")
@register_func(COORDS)
def monotonic_float_coord():
    """Create coordinates which are not evenly sampled."""
    ar = np.cumsum(np.abs(random_state.rand(100)))
    return get_coord(data=ar)


@pytest.fixture(scope="session")
@register_func(COORDS)
def reverse_monotonic_float_coord():
    """Create coordinates which are evenly sampled."""
    ar = np.cumsum(np.abs(random_state.rand(100)))[::-1]
    return get_coord(data=ar)


@pytest.fixture(scope="session")
@register_func(COORDS)
def monotonic_datetime_coord():
    """Create coordinates which are evenly sampled."""
    ar = np.cumsum(np.abs(random_state.rand(100) * 1_000))
    return get_coord(data=dc.to_datetime64(ar))


@pytest.fixture(scope="session")
@register_func(COORDS)
def reverse_monotonic_datetime_coord():
    """Create coordinates which are evenly sampled."""
    ar = -np.cumsum(np.abs(random_state.rand(100) * 1_000))
    return get_coord(data=dc.to_datetime64(ar))


@pytest.fixture(scope="session")
@register_func(COORDS)
def random_coord():
    """Create coordinates which are evenly sampled."""
    ar = random_state.rand(100) * 1_000
    return get_coord(data=ar)


@pytest.fixture(scope="session")
@register_func(COORDS)
def random_date_coord():
    """Create coordinates which are evenly sampled."""
    ar = random_state.rand(100) * 1_000
    return get_coord(data=dc.to_datetime64(ar))


# This is a "special case" so it isnt aggregated into meta fixture.
@pytest.fixture(scope="class")
def basic_non_coord():
    """Create a simple non coord."""
    out = get_coord(data=10)
    return out


@pytest.fixture(scope="session", params=COORDS)
def coord(request) -> BaseCoord:
    """Meta-fixture for returning all coords."""
    return request.getfixturevalue(request.param)


@pytest.fixture(scope="session", params=COORDS)
def long_coord(coord) -> BaseCoord:
    """Meta-fixture for returning all coords with len > 7."""
    if len(coord) < 7:
        pytest.skip("Only coords with len 3 or more used.")
    return coord


@pytest.fixture()
def degenerate_time_coord():
    """Return a simple degenerate coord."""
    return get_coord(data=10, dtype="datetime64[ns]")


@pytest.fixture(scope="session")
def two_d_coord():
    """Return a 2D coordinate."""
    ar = np.ones((10, 10))
    return get_coord(data=ar, units="m/s")


def assert_value_in_one_step(coord, index, value, greater=True):
    """Ensure value at index is within one step of value."""
    # reverse greater for reverse monotonic
    coord_value = coord.values[index]
    if greater:
        assert value <= coord_value
    else:
        assert value >= coord_value
    # next we ensure value is in the correct index. This is a little
    # bit complicated by reverse order coords.
    val1 = coord.values[index + 1]
    val2 = coord.values[index - 1]
    assert (val1 <= value <= val2) or (val2 <= value <= val1)


class TestBasics:
    """A suite of basic tests for coordinates."""

    def test_coord_init(self, coord):
        """Simply run to insure all coords initialize."""
        assert isinstance(coord, BaseCoord)

    def test_bad_init(self):
        """Ensure no parameters raises error."""
        with pytest.raises(CoordError):
            get_coord()

    def test_value(self, coord):
        """All coords should return an array with the values attr."""
        ar = coord.values
        assert isinstance(ar, np.ndarray)

    def test_len_one_array(self):
        """Ensure one length array returns coord."""
        ar = np.array([10])
        coord = get_coord(data=ar)
        assert isinstance(coord, BaseCoord)
        assert len(coord) == 1

    def test_empty_array(self):
        """An empty array should also be a valid coordinate."""
        ar = np.array([])
        coord = get_coord(data=ar)
        assert isinstance(coord, BaseCoord)
        assert len(coord) == 0

    def test_coord_input(self, coord):
        """A coordinate should be valid input."""
        assert get_coord(data=coord) == coord

    def test_cant_add_extra_fields(self, evenly_sampled_coord):
        """Ensure coordinates are immutable."""
        with pytest.raises(ValueError, match="Instance is frozen"):
            evenly_sampled_coord.bob = 1

    def test_sort(self, coord):
        """Ensure every coord can be sorted."""
        if coord.degenerate or len(coord.shape) > 1:
            return  # no need to test degenerate coord or multidim
        data = coord.data
        new, indexer = coord.sort()
        assert new.dtype == coord.dtype
        assert new.sorted
        assert not new.degenerate
        assert not new.reverse_sorted
        assert len(new) == len(coord)
        assert data[indexer] is not None

    def test_reverse_sort(self, coord):
        """Ensure every coord can be reverse sorted."""
        if coord.degenerate or len(coord.shape) > 1:
            return  # no need to test degenerate coord or multidim
        data = coord.data
        new, indexer = coord.sort(reverse=True)
        assert new.dtype == coord.dtype
        assert not new.sorted
        assert new.reverse_sorted
        assert len(new) == len(coord)
        assert data[indexer] is not None

    def test_immutable(self, evenly_sampled_coord):
        """Fields can't change once created."""
        with pytest.raises((TypeError, ValueError), match="Instance is frozen"):
            evenly_sampled_coord.start = 10

    def test_values_immutable(self, coord):
        """Values should all be immutable arrays."""
        with pytest.raises(ValueError, match="assignment destination is read-only"):
            coord.data[0] = coord.data[1]

    def test_str(self, coord):
        """All coords should be convertible to str."""
        out = str(coord)
        assert isinstance(out, str)

    def test_rich(self, coord):
        """Each coord should have nice rich printing."""
        out = coord.__rich__()
        assert isinstance(out, rich.text.Text)

    def test_snap(self, coord):
        """Tests for snapping coordinates."""
        out = coord.snap()
        assert isinstance(out, BaseCoord)
        assert out.shape == coord.shape
        assert out.size == coord.size
        # sort order should stay the same
        if coord.reverse_sorted:
            assert out.reverse_sorted
        if coord.sorted:
            assert out.sorted

    def test_default_units(self):
        """Default units should be None."""
        out = get_coord(data=np.arange(10))
        assert out.units is None

    def test_min_max_align_with_array(self, coord):
        """Min/max values should match the same operation on data."""
        if coord.degenerate:
            return
        values = coord.values
        assert coord.min() == np.min(values)
        assert coord.max() == np.max(values)
        assert coord.max() > coord.min()

    def test_can_pickle(self, coord):
        """Ensure each coord can be pickled/unpickled, both should be equal."""
        len(coord)
        bio = BytesIO()
        pickle.dump(coord, bio)
        bio.seek(0)
        new = pickle.load(bio)
        assert new == coord

    def test_like_coords_have_s_units(self, coord):
        """Time like coordinates should have units of s set automatically."""
        if dtype_time_like(coord.dtype):
            assert coord.units == get_quantity("s")

    def test_convert_simple_units(self, coord):
        """Ensure units can be converted/set."""
        # time always units of s
        if dtype_time_like(coord.dtype):
            pytest.skip("Time units cannot be converted.")
        out_m = coord.convert_units("m")
        assert dc.get_unit(out_m.units) == dc.get_unit("m")
        out_mm = out_m.convert_units("mm")
        assert dc.get_unit(out_mm.units) == dc.get_unit("mm")
        assert len(out_mm) == len(out_m)
        assert np.allclose(out_mm.values / 1000, out_m.values)

    def test_convert_offset_units(self):
        """Ensure units can be converted/set for offset units."""
        array = np.arange(10)
        f_array = array * (9 / 5) + 32.0
        coord = get_coord(data=array, units="degC")
        out = coord.convert_units("degF")
        assert out.shape == f_array.shape
        assert np.allclose(f_array, out.values)

    def test_unit_str(self, evenly_sampled_coord):
        """Ensure the unit string returns a string."""
        coord = evenly_sampled_coord.set_units("m/s")
        assert isinstance(coord.unit_str, str)

    def test_out_of_range_raises(self, evenly_sampled_coord):
        """Accessing a value out of the range of array should raise."""
        with pytest.raises(IndexError, match="exceeds coord length"):
            _ = evenly_sampled_coord[len(evenly_sampled_coord)]

    def test_to_array(self, coord):
        """Ensure np.array converts coord to an array."""
        ar = np.array(coord)
        assert isinstance(ar, np.ndarray)
        assert len(ar) == len(coord)

    def test_bad_range_step(self):
        """Ensure start, stop, step are consistent. See #321."""
        msg = "Sign of step must match"
        with pytest.raises(ValidationError, match=msg):
            get_coord(start=1, stop=5, step=-1)
        with pytest.raises(ValidationError, match=msg):
            get_coord(start=1, stop=-5, step=1)

    def test_both_values_and_data_raises(self):
        """Cannot specify both values and data."""
        data = np.empty(10)
        msg = "Cannot specify both data and values"
        with pytest.raises(CoordError, match=msg):
            get_coord(data=data, values=data)

    def test_coord_range(self, monotonic_float_coord):
        """Ensure that coord_range raises an error for not evenly sampled patches."""
        msg = "has to be called on an evenly sampled"
        with pytest.raises(CoordError, match=msg):
            monotonic_float_coord.coord_range()

    def test_get_coord_datetime(self):
        """Ensure get_coord accepts a datetime object. See #467."""
        d1 = datetime.datetime.fromisoformat("2017-09-18T01")
        d2 = datetime.datetime.fromisoformat("2017-09-18T02")
        step = datetime.timedelta(minutes=1)
        coord = get_coord(start=d1, stop=d2, step=step)
        assert isinstance(coord, BaseCoord)
        assert str(d1) == str(coord.min())
        assert str(d2) == str(coord.max() + coord.step)


class TestCoordSummary:
    """tests for converting to and from summary coords."""

    cast_data_list: ClassVar = [
        {"min": 400.0, "step": 1.0, "units": "", "max": 1000.0},
        {"min": 0, "max": 10},
        {"min": dc.to_datetime64(10), "max": dc.to_datetime64(100)},
        {"min": dc.to_timedelta64(100), "max": dc.to_timedelta64(200)},
    ]

    @pytest.fixture(scope="session")
    def summary(self, coord) -> CoordSummary:
        """Convert each coord to a summary."""
        return coord.to_summary()

    def test_dtype_consistent(self, summary, coord):
        """Ensure the datatype is preserved."""
        dtype = np.dtype(summary.dtype)
        min_ar = np.array(summary.min)
        max_ar = np.array(summary.max)
        assert np.issubdtype(min_ar.dtype, dtype)
        assert np.issubdtype(max_ar.dtype, dtype)

    def test_to_summary(self, coord):
        """Ensure all coords can be converted to a summary."""
        out = coord.to_summary()
        assert isinstance(out, CoordSummary)

    def test_coord_range_round_trip(self, coord):
        """Coord ranges should round-trip to summaries and back."""
        if not coord.evenly_sampled:
            return
        summary = coord.to_summary()
        back = summary.to_coord()
        assert back == coord

        if not back.to_summary() == summary:
            coord.to_summary()

        assert back.to_summary() == summary

    def test_to_summary_raises(self, random_coord):
        """Ensure to_summary raises if not evenly sampled."""
        match = "Cannot convert summary which is not evenly sampled"
        with pytest.raises(CoordError, match=match):
            random_coord.to_summary().to_coord()

    @pytest.mark.parametrize("data", cast_data_list)
    def test_dtype_inferred(self, data):
        """Ensure the dtypes are correctly determined if not specified."""
        out = CoordSummary(**data)
        assert np.dtype(out.dtype) == np.dtype(type(data["min"]))


class TestGetSliceTuple:
    """Tests for getting slice tuples."""

    def test_get_slice_tuple(self, coord):
        """Basic tests for range of coords."""
        start = coord.min()
        end = coord.max()
        out = coord.get_slice_tuple((start, end))
        assert out == (start, end)
        assert coord.get_slice_tuple((start, None)) == (start, None)
        assert coord.get_slice_tuple((None, end)) == (None, end)
        # if units aren't none, ensure they work.
        if not (unit := coord.units):
            return
        # since we can't multiply dt by quant, convert to float
        if dtype_time_like(coord):
            start, end = to_float(start), to_float(end)
        start_unit = start * get_quantity(unit)
        end_unit = end * get_quantity(unit)
        # check that first slice is close to original
        sliced = coord.get_slice_tuple((start_unit, None))
        lims = (coord.min(), None)
        assert all_close([to_float(lims[0])], [to_float(sliced[0])])
        # same for second slice.
        lims = to_float(np.array((coord.min(), coord.max())))
        sliced = to_float(np.array(coord.get_slice_tuple((start_unit, end_unit))))
        assert all_close(lims, sliced)

    def test_get_slice_range_bad_values(self, evenly_sampled_coord):
        """Ensure bad values raise helpful error."""
        with pytest.raises(ParameterError, match="must be a length 2"):
            evenly_sampled_coord.get_slice_tuple((1, 2, 3))

    def test_slice_with_step_raises(self, evenly_sampled_coord):
        """A slice with a step shouldn't work."""
        match = "Step not supported"
        with pytest.raises(ParameterError, match=match):
            evenly_sampled_coord.select(slice(1, 10, 2))

    def test_slice_works_as_tuple(self, evenly_sampled_coord):
        """Ensure slice works like tuple."""
        coord = evenly_sampled_coord
        vmin, vmax = coord.min(), coord.max()
        sli = slice(vmin, vmax + 1)
        out_sli = evenly_sampled_coord.get_slice_tuple(sli)
        assert out_sli == (None, None) or out_sli == (0, len(coord))


class TestSelect:
    """Generic tests for selecting values from coords."""

    def test_select_end_end_time(self, coord):
        """Ensure when time range is == (end, end) that dim has len 1."""
        out = coord.select((coord.max(), coord.max()))[0]
        assert len(out) == 1

    def test_intra_sample_select(self, coord):
        """
        Selecting ranges that fall within samples should become de-generate.
        This is consistent with pandas indices.
        """
        values = coord.values
        # get value between first/second sample
        arg = values[0] + (values[1] - values[0]) / 2
        assert arg not in np.unique(values)
        out, indexer = coord.select((arg, arg))
        assert out.degenerate
        assert np.size(coord.data[indexer]) == 0

    def test_select_start_start_time(self, coord):
        """Ensure when time range is == (start, start) that dim has len 1."""
        out = coord.select((coord.min(), coord.min()))[0]
        assert len(out) == 1

    def test_select_inclusive(self, long_coord):
        """Ensure selecting is inclusive on both ends."""
        values = np.sort(long_coord.values)
        value_set = set(values)
        assert len(values) > 7
        args = (values[3], values[-3])
        new, _ = long_coord.select(args)
        # Check min and max values conform to args
        new_values = new.values
        new_min, new_max = np.min(new_values), np.max(new_values)
        assert (new_min >= args[0]) or np.isclose(new_min, args[0])
        assert (new_max <= args[1]) or np.isclose(new_max, args[1])
        # There can often be a bit of "slop" in float indices. We may
        # need to rethink this but this is good enough for now.
        if not np.issubdtype(new.dtype, np.float64):
            assert {values[3], values[-3]}.issubset(value_set)

    def test_select_bounds(self, long_coord):
        """Ensure selecting bounds are honored."""
        values = np.sort(long_coord.values)
        # get a little more than and little less than 2 and -2 index vals
        val1 = values[2] + (values[3] - values[2]) / 2
        val2 = values[-2] - (values[-2] - values[-3]) / 2
        new, _ = long_coord.select((val1, val2))
        new_values = new.values
        new_min, new_max = np.min(new_values), np.max(new_values)
        assert (new_min >= val1) or np.isclose(new_min, val1)
        assert (new_max <= val2) or np.isclose(new_max, val2)
        if not np.issubdtype(new.dtype, np.float64):
            assert set(values).issuperset(set(new.values))

    def test_select_out_of_bounds_too_early(self, coord):
        """Applying a select out of bounds (too early) should return partial."""
        diff = (coord.max() - coord.min()) / (len(coord) - 1)
        # get a range which is for sure before data.
        v1 = coord.min() - np.abs(100 * diff)
        v2 = v1 + np.abs(30 * diff)
        # it should return a degenerate because range not contained by coord.
        new, indexer = coord.select((v1, v2))
        assert new.degenerate
        assert np.size(coord.data[indexer]) == 0
        # Same thing if end time is too early
        new, indexer = coord.select((None, v2))
        assert new.degenerate
        assert np.size(coord.data[indexer]) == 0
        # but this should be fine
        new = coord.select((v1, None))[0]
        assert new == coord

    def test_select_out_of_bounds_too_late(self, coord):
        """Applying a select out of bounds (too late) should return partial."""
        diff = (coord.max() - coord.min()) / (len(coord) - 1)
        # get a range which is for sure after data.
        v1 = coord.max() + np.abs(100 * diff)
        v2 = v1 + np.abs(30 * diff)
        # it should return a degenerate since out of range
        new, indexer = coord.select((v1, v2))
        assert new.degenerate
        assert np.size(coord.data[indexer]) == 0
        # Same thing if start time is too late
        new, indexer = coord.select((v1, None))
        assert new.degenerate
        assert np.size(coord.data[indexer]) == 0
        assert coord.select((None, v2))[0] == coord

    def test_wide_select_bounds(self, coord):
        """Wide (lt and gt limits) select bounds should be fine."""

        def _is_equal(coord1, coord2, indexer):
            """Assert coord and indexer are equal."""
            assert coord1 == coord2
            if isinstance(indexer, slice):
                assert indexer == slice(None, None)
            elif isinstance(indexer, np.ndarray):
                assert np.all(indexer)

        diff = (coord.max() - coord.min()) / (len(coord) - 1)
        v1 = coord.min() - 10 * diff
        v2 = coord.max() + 10 * diff
        out, slice_thing = coord.select((v1, v2))
        _is_equal(coord, out, slice_thing)
        out, slice_thing = coord.select((None, v2))
        _is_equal(coord, out, slice_thing)
        out, slice_thing = coord.select((v1, None))
        _is_equal(coord, out, slice_thing)

    def test_select_ordered_coords(self, coord):
        """Basic select tests all coords should pass."""
        if not (coord.sorted or coord.reverse_sorted):
            return
        assert len(coord.values) == len(coord)
        value1 = coord.values[10]
        new, sliced = coord.select((value1, ...))
        # this accounts for reverse sorted coords
        slice_value = sliced.start if coord.sorted else sliced.stop - 1
        assert_value_in_one_step(coord, slice_value, value1, greater=True)
        assert new.values is not coord.values, "new data should be made"
        assert slice_value == 10
        # now test exact value
        value2 = coord.values[20]
        new, sliced = coord.select((value2, ...))
        slice_value = sliced.start if coord.sorted else sliced.stop - 1
        assert_value_in_one_step(coord, slice_value, value2, greater=True)
        assert slice_value == 20
        # test end index
        new, sliced = coord.select((..., value1))
        assert sliced.stop == 11 if new.sorted else sliced.start == 10
        new, sliced = coord.select((..., value2))
        assert sliced.stop == 21 if new.sorted else sliced.start == 20
        # test range
        new, sliced = coord.select((value1, value2))
        assert len(new.values) == 11 == len(new)
        assert sliced == slice(10, 21)

    def test_select_relative(self, coord):
        """Test relative select from start/end."""
        dist = (coord.max() - coord.min()) / 4
        abs1 = coord.min() + dist
        abs2 = coord.max() - dist
        out1, ind_1 = coord.select((abs1, abs2))
        out2, ind_2 = coord.select((dist, -dist), relative=True)
        assert np.all(out1 == out2)
        if isinstance(ind_1, np.ndarray):
            assert np.all(ind_1 == ind_2)
        else:
            assert ind_1 == ind_2

    def test_select_changes_len(self, long_coord):
        """Ensure select changes the length of the coordinate."""
        data = long_coord.data
        new, dslice = long_coord.select((data[2], data[-2]))
        if hasattr(dslice, "stop") and hasattr(dslice, "start"):
            new_len = dslice.stop - dslice.start
        else:
            new_len = (
                len(dslice) if np.issubdtype(dslice.dtype, np.integer) else dslice.sum()
            )
        assert len(new) == new_len

    def test_select_sub_array(self, long_coord):
        """A sub-array should allow indexing."""
        values = long_coord.values
        sub = values[1:-1]
        out, reduction = long_coord.select(sub)
        assert reduction.sum() == len(sub)

    def test_overlapping_array(self, long_coord):
        """Ensure an overlapping array works as well."""
        # Create an array which likely has values not contained in the original.
        values1 = long_coord.values[: len(long_coord) // 2]
        values2 = long_coord._get_compatible_value(-to_float(values1))
        values = np.concatenate([values1, values2])
        # Ensure this still gets some output.
        in_coord = np.isin(long_coord.values, values)
        out, reduction = long_coord.select(values)
        assert in_coord.sum() == len(out) == reduction.sum()

    def test_duplicate_array_samples(self, long_coord):
        """Ensure duplicate do nothing."""
        inds = np.array([0, 0, 0])
        coord, reduction = long_coord.select(inds, samples=True)
        assert len(coord) == len(np.unique(inds))
        assert np.all(coord.values == coord.values[0])

    def test_non_integer_array_with_samples_raises(self, evenly_sampled_coord):
        """Samples argument should require integer arrays."""
        vals = np.array([1.01, 2.0, 3.0])
        msg = "requires integer dtype"
        with pytest.raises(CoordError, match=msg):
            evenly_sampled_coord.select(vals, samples=True)

    def test_duplicate_array_values(self, long_coord):
        """Ensure duplicate values don't cause duplicates in array."""
        second_value = long_coord.values[1]
        array = np.array([second_value, second_value])
        coord, reduction = long_coord.select(array)
        assert len(coord) == len(np.unique(array))
        assert np.all(array == second_value)

    def test_monotonic_select_by_samples(self, monotonic_float_coord):
        """Ensure we can select by samples with monotonic coord."""
        i1, i2 = 15, 50
        values = monotonic_float_coord.values
        out = monotonic_float_coord.select((i1, i2), samples=True)[0]
        assert np.all(values[i1:i2] == out.values)

    def test_values_not_in_coord(self, long_coord):
        """Ensure using values not in coord results in an empty coordinate."""
        values1 = long_coord.values[: len(long_coord) // 2]
        rand = random_state.rand(len(values1))
        # try to make values that won't be in the coordinate.
        values2 = long_coord._get_compatible_value(-to_float(values1) + rand)
        # if by some chance there are any overlaps just bail out.
        if np.any(np.isin(values2, long_coord.data)):
            return
        out, red = long_coord.select(values2)
        assert out.shape == red.sum()
        assert len(out) == 0

    def test_index_not_in_coord(self, long_coord):
        """
        Ensure using samples array with values not in coord results in
        an empty coordinate.
        """
        bad_indices = np.arange(len(long_coord)) + len(long_coord)
        out, red = long_coord.select(bad_indices, samples=True)
        assert out.shape == red.sum()
        assert len(out) == 0

    def test_array_boolean_all_true(self, long_coord):
        """Ensure a boolean array works to filter values."""
        bools = np.ones(len(long_coord), dtype=np.bool_)
        out, red = long_coord.select(bools)
        assert len(out) == bools.sum()
        assert np.all(red == bools)

    def test_array_boolean_all_false(self, long_coord):
        """Ensure an array of all false values works to filter coord."""
        bools = np.zeros(len(long_coord), dtype=np.bool_)
        out, red = long_coord.select(bools)
        assert len(out) == bools.sum()
        assert np.all(red == bools)

    def test_mix_bools(self, long_coord):
        """Ensure mixed booleans also work."""
        bools = np.zeros(len(long_coord), dtype=np.bool_)
        bools[::2] = True
        out, red = long_coord.select(bools)
        assert len(out) == bools.sum()
        assert np.all(red == bools)

    def test_bad_range_raises(self, random_coord):
        """Ensure a non-tuple or slice raises an Error."""
        bad_vals = [1, 10.0, "hey"]
        for val in bad_vals:
            msg = "Range values must be"
            with pytest.raises(ParameterError, match=msg):
                random_coord.select(val)

    def test_select_samples_no_int_raises(self, random_coord):
        """An error is raised when samples=True and non int values are used."""
        expected = "When samples=True"
        with pytest.raises(ParameterError, match=expected):
            random_coord.select((1, 1.2), samples=True)


class TestOrder:
    """Tests for ordering coordinates."""

    @pytest.fixture(scope="class")
    def coord_with_duplicates(self):
        """Create a coordinate with duplicate values."""
        ar = np.array([0, 0, 1, 1, 1, 2, 2, 2, 2])
        return get_coord(data=ar)

    def test_select_sub_array(self, long_coord):
        """A sub-array should allow indexing."""
        values = long_coord.values
        sub = values[1:-1]
        out, reduction = long_coord.order(sub)
        assert len(out) == len(values[reduction])

    def test_overlapping_array(self, long_coord):
        """Ensure an overlapping array works as well."""
        # Create an array which likely has values not contained in the original.
        values1 = long_coord.values[: len(long_coord) // 2]
        values2 = long_coord._get_compatible_value(-to_float(values1))
        values = np.concatenate([values1, values2])
        # Ensure this still gets some output.
        in_coord = np.isin(values, long_coord.values)
        out, reduction = long_coord.order(values)
        assert in_coord.sum() == len(out) == len(long_coord.values[reduction])

    def test_duplicate_array_samples(self, long_coord):
        """Ensure duplicate indices cause duplicates in array."""
        inds = np.array([0, 0, 0])
        coord, reduction = long_coord.order(inds, samples=True)
        assert len(coord) == len(inds)
        assert np.all(coord.values == coord.values[0])

    def test_non_integer_array_with_samples_raises(self, evenly_sampled_coord):
        """Samples argument should require integer arrays."""
        vals = np.array([1.01, 2.0, 3.0])
        msg = "requires integer dtype"
        with pytest.raises(CoordError, match=msg):
            evenly_sampled_coord.select(vals, samples=True)

    def test_duplicate_array_values(self, long_coord):
        """Ensure duplicate values cause duplicates in array."""
        second_value = long_coord.values[1]
        array = np.array([second_value, second_value])
        coord, reduction = long_coord.order(array)
        assert len(coord) == len(array)
        assert np.all(array == second_value)

    def test_sorted_by_values(self, long_coord):
        """Ensure the output is sorted by values given to select."""
        vals = long_coord.values[np.array([4, 2, 1, 3])]
        coord, reduction = long_coord.order(vals)
        assert vals.shape == coord.shape
        assert np.all(vals == coord.values)

    def test_values_no_in_coord(self, long_coord):
        """Ensure using values not in coord results in an empty coordinate."""
        values1 = long_coord.values[: len(long_coord) // 2]
        rand = random_state.rand(len(values1))
        # try to make values that won't be in the coordinate.
        values2 = long_coord._get_compatible_value(-to_float(values1) + rand)
        # if by some chance there are any overlaps just bail out.
        if np.any(np.isin(values2, long_coord.data)):
            return
        out, red = long_coord.order(values2)
        assert out.shape == red.shape
        assert len(out) == 0

    def test_index_not_in_coord(self, long_coord):
        """
        Ensure using samples array with values not in coord results in
        an empty coordinate.
        """
        bad_indices = np.arange(len(long_coord)) + len(long_coord)
        out, red = long_coord.order(bad_indices, samples=True)
        assert out.shape == red.shape
        assert len(out) == 0

    def test_duplicate_coord_single_value(self, coord_with_duplicates):
        """Ensure order works with duplicate coordinates and single value."""
        old_values = coord_with_duplicates.values
        out, inds = coord_with_duplicates.order(np.asarray(1))
        assert np.all(out.values == 1)
        assert np.all(old_values[inds] == 1)
        assert len(out) == (out.values == 1).sum()

    def test_duplicate_coord_multiple_values(self, coord_with_duplicates):
        """Ensure order works with duplicate coordinates and values."""
        # also testing out of order indices
        old_values = coord_with_duplicates.values
        to_find = [1, 0]
        out, inds = coord_with_duplicates.order(np.asarray(to_find))
        assert set(out.values) == {1, 0}
        assert len(out) == np.isin(old_values, to_find).sum()
        expected = np.concatenate([np.where(old_values == x)[0] for x in to_find])
        assert np.all(inds == expected)

    def test_duplicate_coord_duplicate_values(self, coord_with_duplicates):
        """Test duplicate coord values and duplicate search values."""
        old_values = coord_with_duplicates.values
        to_find = [0, 1, 1]
        expected_len = sum((old_values == x).sum() for x in to_find)
        out, inds = coord_with_duplicates.order(to_find)
        assert len(out) == expected_len

    def test_order_samples_non_int_aray(self, evenly_sampled_coord):
        """Ensure an error is raised if non-int array is used with samples."""
        msg = "requires integer dtype"
        bad_array = np.array([1.22, 2.33])
        with pytest.raises(CoordError, match=msg):
            evenly_sampled_coord.order(bad_array, samples=True)


class TestEqual:
    """Tests for comparing coord equality."""

    def test_different_units_not_equal(self, coord):
        """When different units are set the coords should not be equal."""
        if coord.units == "furlongs":
            return
        if dtype_time_like(coord.dtype):
            pytest.skip("Cant set time-like units")
        new = coord.set_units("furlongs")
        assert new != coord

    def test_compare_non_coord_coord_range(self, evenly_sampled_coord):
        """Ensure non-coords compare false."""
        assert evenly_sampled_coord is not None
        assert not evenly_sampled_coord == {1, 2}

    def test_compare_non_coord_coord_array(self, random_coord):
        """Ensure non-coords compare false."""
        assert random_coord is not None
        assert not random_coord == {1, 2}


class TestApproxEqual:
    """Tests for coordinates being approximately equal."""

    def test_non_coord(self, basic_non_coord):
        """Ensure non-coords of same length are equal."""
        assert basic_non_coord.approx_equal(basic_non_coord)

    def test_normal_coord(self, evenly_sampled_coord):
        """Ensure evenly sampled coords are equal."""
        assert evenly_sampled_coord.approx_equal(evenly_sampled_coord)

    def test_different_shape(self, evenly_sampled_coord):
        """Coords of different shapes !=."""
        sub, _ = evenly_sampled_coord.select(0, 2, samples=True)
        assert not sub.approx_equal(evenly_sampled_coord)
        assert not evenly_sampled_coord.approx_equal(sub)

    def test_different_values_ne(self, evenly_sampled_float_coord_with_units):
        """Ensure very different value coords are not approx equal."""
        coord = evenly_sampled_float_coord_with_units
        out = coord.update(data=coord.data * 1_000_000)
        assert not out.approx_equal(coord)

    def test_close_values_equal(self, evenly_sampled_float_coord_with_units):
        """Ensure very close value coords are equal."""
        coord = evenly_sampled_float_coord_with_units
        out = coord.update(data=coord.data * 0.99999999999)
        assert not np.all(out.data == coord.data)
        assert out.approx_equal(coord)


class TestCoordRange:
    """Tests for coords from array."""

    def test_init_array(self, evenly_sampled_coord):
        """Ensure the array can be initialized."""
        assert evenly_sampled_coord.step == 1.0
        assert evenly_sampled_coord.units is None

    def test_values(self, evenly_sampled_coord):
        """Ensure we can get an array with values attribute."""
        vals = evenly_sampled_coord.values
        assert len(vals) == len(evenly_sampled_coord)

    def test_set_units(self, evenly_sampled_coord):
        """Ensure units can be set."""
        out = evenly_sampled_coord.set_units("m")
        assert out.units == get_quantity("m")
        assert out is not evenly_sampled_coord

    def test_set_quantity(self, evenly_sampled_coord):
        """Ensure units can be set."""
        out = evenly_sampled_coord.set_units("10 m")
        assert out.units == get_quantity("10 m")
        assert out is not evenly_sampled_coord

    def test_dtype(self, evenly_sampled_coord, evenly_sampled_date_coord):
        """Ensure datatypes are sensible."""
        dtype1 = evenly_sampled_coord.dtype
        assert np.issubdtype(dtype1, np.int_)
        dtype2 = evenly_sampled_date_coord.dtype
        assert np.issubdtype(dtype2, np.datetime64)

    def test_select_tuple_ints(self, evenly_sampled_coord):
        """Ensure a tuple works as a limit."""
        assert evenly_sampled_coord.select((50, None))[1] == slice(50, None)
        evenly_sampled_coord.select((0, None))
        assert evenly_sampled_coord.select((0, None))[1] == slice(None, None)
        assert evenly_sampled_coord.select((-10, None))[1] == slice(None, None)
        assert evenly_sampled_coord.select((None, None))[1] == slice(None, None)
        assert evenly_sampled_coord.select((None, 1_000_000))[1] == slice(None, None)

    def test_identity_slice_ints(self, evenly_sampled_coord):
        """Ensure slice with exact start/end gives same coord."""
        coord = evenly_sampled_coord
        new, sliced = coord.select((coord.start, coord.stop))
        assert new == coord

    def test_identity_slice_floats(self, evenly_sampled_float_coord_with_units):
        """Ensure slice with exact start/end gives same coord."""
        coord = evenly_sampled_float_coord_with_units
        new, sliced = coord.select((coord.start, coord.stop))
        assert new == coord

    def test_float_basics(self):
        """Ensure floating point range coords work."""
        ar = np.arange(1, 100, 1 / 3)
        out = get_coord(data=ar, units="m")
        assert len(out) == len(ar)
        assert np.allclose(out.values, ar)

    def test_select_units(self, evenly_sampled_float_coord_with_units):
        """Ensure units can work in select."""
        coord = evenly_sampled_float_coord_with_units
        value_ft = 100
        value_m = value_ft * 0.3048
        new, sliced = coord.select((value_ft * dc.get_unit("ft"), ...))
        # ensure value is surrounded.
        assert new.start + new.step >= value_m
        assert new.start - new.step <= value_m
        # units should also not change
        assert new.units == coord.units

    def test_set_convert_units(self, coord):
        """Simple test for setting and converting units."""
        if dtype_time_like(coord.dtype):
            pytest.skip("Cant set time-like units")
        factor = get_quantity("m").to(get_quantity("ft")).magnitude
        c1 = coord.set_units("m")
        c2 = c1.convert_units("ft")
        assert c2.units == get_quantity("ft")
        if c2.degenerate:
            return
        values1 = c1.values
        values2 = c2.values
        # for datetime and timedelta no conversion should happen
        if is_datetime64(values1[0]) or is_timedelta64(values1[0]):
            assert np.all(np.equal(values1, values2))
        else:
            np.allclose(values1, values2 / factor)

    def test_select_date_string(self, evenly_sampled_date_coord):
        """Ensure string selection works with datetime objects."""
        coord = evenly_sampled_date_coord
        date_str = "1970-01-01T12"
        new, sliced = coord.select((date_str, ...))
        datetime = dc.to_datetime64(date_str)
        assert new.start + new.step >= datetime
        assert new.start - new.step <= datetime

    def test_sort(self, evenly_sampled_coord):
        """Ensure sort returns equal coord."""
        out, _ = evenly_sampled_coord.sort()
        assert out == evenly_sampled_coord

    def test_update_limits_too_many_params(self, evenly_sampled_coord):
        """Update limits should raise if too many parameters are specified."""
        with pytest.raises(ValueError, match="At most two parameters"):
            evenly_sampled_coord.update_limits(1, 10, 1)

    def test_update_limits_start(self, evenly_sampled_coord):
        """Ensure start can be updated."""
        new_start = evenly_sampled_coord.start + 2 * evenly_sampled_coord.step
        new = evenly_sampled_coord.update_limits(min=new_start)
        assert len(new) == len(evenly_sampled_coord)
        assert new.step == evenly_sampled_coord.step
        assert new.start == new_start

    def test_update_limits_end(self, evenly_sampled_coord):
        """Ensure end can be updated."""
        new_stop = evenly_sampled_coord.start - 2 * evenly_sampled_coord.step
        step = evenly_sampled_coord.step
        new = evenly_sampled_coord.update_limits(max=new_stop - step)
        assert len(new) == len(evenly_sampled_coord)
        assert new.step == evenly_sampled_coord.step
        assert new.stop == new_stop

    def test_update_limits_step(self, evenly_sampled_coord):
        """Ensure the step can be updated."""
        coord = evenly_sampled_coord
        new_step = coord.step * 2
        new = coord.update_limits(step=new_step)
        assert np.all(new.values == coord.values * 2)

    def test_update_limits_start_stop(self, evenly_sampled_coord):
        """Ensure both start/stop can be updated."""
        coord = evenly_sampled_coord
        start, stop = coord.values[2], coord.values[-2]
        new = coord.update_limits(min=start, max=stop)
        assert len(new) == len(coord)
        assert new.start == start
        assert new.stop == stop

    def test_update_limits_no_args(self, evenly_sampled_coord):
        """Ensure both start/stop can be updated."""
        coords = evenly_sampled_coord
        assert coords == coords.update_limits()

    def test_update_limits_timedelta64_different_precision(self):
        """Update limits with datetime64 that have different precisions."""
        start, stop = np.datetime64("2023-01-01"), np.datetime64("2023-01-01T01")
        step = np.timedelta64(60, "s")
        coord = get_coord(start=start, stop=stop, step=step)
        # change maximum value, keeping length the same and chaning step
        coord_new = coord.update_limits(max=stop + 10 * step)
        assert coord_new.max() == stop + 10 * step

    def test_test_select_end_floats(self, evenly_sampled_float_coord_with_units):
        """Ensure we can select right up to the end of the array."""
        coord = evenly_sampled_float_coord_with_units
        new_coord, out = coord.select((coord.min(), coord.max()))
        assert len(new_coord) == len(coord)
        assert np.allclose(new_coord.values, coord.values)

    def test_empty(self, coord):
        """Ensure coords can be emptied out."""
        new = coord.empty()
        assert new.degenerate
        assert new.dtype == coord.dtype

    def test_init_length_one(self):
        """Ensure len 1 coord can be inited provided step is supplied."""
        time = dc.to_datetime64(["2020-01-01"])
        dt = dc.to_timedelta64(0.09999)
        coord1 = get_coord(start=time[0], stop=time[0] + dt, step=dt)
        coord2 = get_coord(data=time, step=dt)
        assert isinstance(coord1, CoordRange)
        assert coord1 == coord2
        # both length and shape should be 1
        assert len(coord1) == 1
        assert coord1.shape == (1,)

    def test_monotonic_with_sampling(self):
        """Ensure init'ing monotonic array with sampling also works."""
        sample_rate = 1_000
        t_array = np.linspace(0.0, 1, 1000)
        sample_rate = 1 / sample_rate
        out = get_coord(data=t_array, step=sample_rate)
        assert out.shape == out.data.shape == (len(out),)

    def test_properties_coord_range(self, evenly_sampled_coord):
        """Test key properties for evenly sampled."""
        coord = evenly_sampled_coord
        assert coord.evenly_sampled
        assert coord.sorted
        assert not coord.reverse_sorted
        assert not coord.degenerate

    def test_coord_range_len_1(self):
        """Ensure coord range can have step size of 0."""
        new = CoordRange(start=0, stop=0, step=0)
        assert len(new) == 1
        assert np.all(new.values == np.zeros(0))

    def test_reversed_coord(self, evenly_sampled_reversed_coord):
        """Ensure reverse sampling works for evenly sampled coord."""
        coord = evenly_sampled_reversed_coord
        assert isinstance(coord, CoordRange)
        assert coord.evenly_sampled
        assert coord.reverse_sorted
        assert not coord.sorted

    def test_arrange_equiv(self):
        """Ensure np.arange produces same len as get_coord, start, stop, step."""
        start = -120.02095199999974
        stop = 1055.2182894582486
        step = 1.0209523838714072
        ar = np.arange(start, stop, step)
        coord = get_coord(start=start, stop=stop, step=step)
        assert coord.shape == ar.shape
        assert coord.size == ar.size

    def test_unchanged_len(self):
        """Ensure an array converted to Coord range has same len. See #229."""
        t_array = np.linspace(0.0, 1, 44100)
        t_coord = get_coord(data=t_array)
        assert t_array.shape == t_coord.shape
        # now test datetime
        time = dc.to_timedelta64(t_array) + np.datetime64("2020-01-01")
        time_coord = get_coord(data=time)
        assert len(time) == len(time_coord)

    def test_length_with_stop(self):
        """The length argument should work for setting start."""
        out1 = CoordRange(start=0, step=1, stop=10)
        out2 = CoordRange(start=0, step=1, shape=(10,))
        assert out1 == out2

    def test_length_with_start(self):
        """The length argument should work for setting stop."""
        out1 = CoordRange(start=0, step=1, stop=10)
        out2 = CoordRange(stop=10, step=1, shape=(10,))
        assert out1 == out2

    def test_length_with_step(self):
        """The length argument should work for setting step."""
        out1 = CoordRange(start=0, step=1, stop=10)
        out2 = CoordRange(start=0, stop=10, shape=10)
        assert out1 == out2


class TestMonotonicCoord:
    """Tests for monotonic array coords."""

    def test_select_basic(self, monotonic_float_coord):
        """Basic select tests for monotonic array."""
        coord = monotonic_float_coord
        # test start only
        new, sliced = coord.select((5, ...))
        assert_value_in_one_step(coord, sliced.start, 5)
        assert new.min() >= 5
        # test stop only
        new, sliced = coord.select((..., 40))
        assert_value_in_one_step(coord, sliced.stop - 1, 40, greater=False)
        assert new.max() <= 40

    def test_sort(self, monotonic_float_coord):
        """Ensure sort returns equal coord."""
        out, _ = monotonic_float_coord.sort()
        assert out == monotonic_float_coord

    def test_reverse_monotonic(self):
        """Ensure reverse monotonic values can be handled."""
        ar = np.cumsum(np.abs(random_state.rand(100)))[::-1]
        coord = get_coord(data=ar)
        assert np.allclose(coord.values, ar)
        new, sliced = coord.select((ar[10], ar[20]))
        assert len(new) == 11 == len(new.values)
        assert np.allclose(new.values, coord.values[10:21])
        # test edges
        eps = 0.000000001
        new, sliced = coord.select((ar[0] + eps, ar[20] - eps))
        assert sliced.start is None
        # ensure messing with the ends keeps the order
        val1, val2 = ar[10] - eps, ar[20] + eps
        new, sliced = coord.select((val1, val2))
        assert sliced == slice(11, 20)
        assert new[0] <= val1
        assert new[-1] >= val2

    def test_index_with_string(self, monotonic_datetime_coord):
        """Ensure indexing works with date string."""
        coord = monotonic_datetime_coord
        value1 = coord.values[12] + np.timedelta64(1, "ns")
        new, sliced = coord.select((value1, ...))
        assert_value_in_one_step(coord, sliced.start, value1, greater=True)
        assert sliced.start == 13
        # now test exact value
        value2 = coord.values[12]
        new, sliced = coord.select((value2, ...))
        assert_value_in_one_step(coord, sliced.start, value2, greater=True)
        assert sliced.start == 12
        # test end index
        new, sliced = coord.select((..., value1))
        assert sliced.stop == 13
        new, sliced = coord.select((..., value2 - np.timedelta64(1, "ns")))
        assert sliced.stop == 12
        # test range
        new, sliced = coord.select((coord.values[10], coord.values[20]))
        new_vals = new.values
        assert coord.values[10] == new_vals[0]
        assert coord.values[20] == new_vals[-1]
        assert len(new) == 11
        assert slice(10, 21) == sliced

    def test_update_limits_too_many_params(self, monotonic_float_coord):
        """Update limits should raise if too many parameters are specified."""
        coords = monotonic_float_coord
        start, stop = coords.min() + 1, coords.max() - 1
        with pytest.raises(ValueError, match="At most one parameter"):
            coords.update_limits(start, stop)

    def test_update_limits_step(self, monotonic_float_coord):
        """Ensure step can be updated."""
        coord = monotonic_float_coord
        new = coord.update_limits(step=1.5)
        assert np.isclose(new.step, 1.5)
        assert new.min() == coord.min()

    def test_update_limits_end(self, monotonic_float_coord):
        """Ensure end can be updated."""
        coord = monotonic_float_coord
        new_stop = coord.max() - 10
        new = coord.update_limits(max=new_stop)
        assert len(new) == len(coord)
        assert new.max() == new_stop

    def test_update_limits_start(self, monotonic_float_coord):
        """Ensure the start can be updated."""
        coord = monotonic_float_coord
        new_start = coord.min() - 100
        new = coord.update_limits(min=new_start)
        assert len(new) == len(coord)
        assert np.isclose(new.min(), new_start)

    def test_update_limits_no_args(self, monotonic_float_coord):
        """Ensure both start/stop can be updated."""
        coords = monotonic_float_coord
        assert coords == coords.update_limits()

    def test_getitem_slice(self, monotonic_float_coord):
        """Ensure get_item_slice returns another coord."""
        new = monotonic_float_coord[slice(1, -1)]
        assert isinstance(new, monotonic_float_coord.__class__)
        assert (len(new) + 2) == len(monotonic_float_coord)

    def test_wide_filter_doesnt_change_size(self, monotonic_float_coord):
        """Ensure filtering outside data range doesn't change array size."""
        coord = monotonic_float_coord
        lims = coord.limits
        dur = lims[1] - lims[0]
        select_range = (lims[0] - 2 * dur, lims[1] + dur)
        wide_coord, inds = coord.select(select_range)
        # coord should be unchanged
        assert len(wide_coord) == len(coord)
        assert wide_coord == coord
        wide_coord, inds = coord.select((select_range[0], None))
        assert wide_coord == coord
        wide_coord, inds = coord.select((None, select_range[1]))
        assert wide_coord == coord

    def test_properties_mono(self, monotonic_float_coord):
        """Check a few properties for monotonic float coord."""
        coord = monotonic_float_coord
        assert coord.sorted
        assert not coord.evenly_sampled
        assert not coord.reverse_sorted
        assert not coord.degenerate

    def test_properties_reverse_mono(self, reverse_monotonic_float_coord):
        """Check a few properties for reverse monotonic float coord."""
        coord = reverse_monotonic_float_coord
        assert not coord.sorted
        assert not coord.evenly_sampled
        assert coord.reverse_sorted
        assert not coord.degenerate

    def test_max_degenerate(self, evenly_sampled_time_delta_coord):
        """Ensure time delta produces nullish value."""
        empty = evenly_sampled_time_delta_coord.empty()
        assert pd.isnull(empty.max())

    def test_single_value(self):
        """A single value coord should be considered sorted."""
        coord = get_coord(data=np.array([1]))
        assert coord.sorted
        assert not coord.reverse_sorted


class TestNonOrderedArrayCoords:
    """Tests for non-ordered array coords."""

    def test_select(self, random_coord):
        """Ensure selecting returns an ndarray."""
        min_v, max_v = np.min(random_coord.values), np.max(random_coord.values)
        dist = max_v - min_v
        val1, val2 = min_v + 0.2 * dist, max_v - 0.2 * dist
        new, ind_array = random_coord.select((val1, val2))
        assert np.all(new.values >= val1)
        assert np.all(new.values <= val2)

    def test_sort(self, random_coord):
        """Ensure the coord can be ordered."""
        new, ordering = random_coord.sort()
        assert isinstance(new, CoordMonotonicArray)

    def test_snap(self, random_coord):
        """Ensure coords can be snapped to even sampling intervals."""
        out = random_coord.snap()
        assert isinstance(out, CoordRange)

    def test_snap_date(self, random_date_coord):
        """Ensure coords can be snapped to even sampling intervals."""
        out = random_date_coord.snap()
        assert np.issubdtype(out.dtype, np.datetime64)
        assert isinstance(out, CoordRange)

    def test_properties_random(self, random_coord):
        """Test key properties for evenly sampled."""
        coord = random_coord
        assert not coord.evenly_sampled
        assert not coord.sorted
        assert not coord.reverse_sorted
        assert not coord.degenerate


class TestCoordArray:
    """Tests for a coordinate managing an array."""

    def test_empty_2d(self, two_d_coord):
        """Ensure the coord can be emptied along any/all axis."""
        out1 = two_d_coord.empty()
        out2 = two_d_coord.empty(axes=1)
        out3 = two_d_coord.empty(axes=0)
        assert out1.size == out2.size == out3.size == 0


class TestPartialCoord:
    """Tests for the Partial coordinate."""

    def test_init_non_coord(self, basic_non_coord):
        """Ensure a non-coord can be created."""
        assert isinstance(basic_non_coord, CoordPartial)

    def test_str_repr(self, basic_non_coord):
        """Ensure NonCoord has a string repr."""
        out = str(basic_non_coord)
        assert isinstance(out, str)

    def test_update(self, basic_non_coord):
        """Ensure update returns a new non coord."""
        out = basic_non_coord.update(shape=2)
        assert len(out) == 2
        assert isinstance(out, CoordPartial)

    def test_non_coord_eq_self(self, basic_non_coord):
        """Ensure non coords are equal to themselves."""
        assert basic_non_coord == basic_non_coord

    def test_empty_update_equal(self, basic_non_coord):
        """Empty update should produce an equal coord."""
        out = basic_non_coord.update()
        assert out == basic_non_coord

    def test_bad_select_raises(self, basic_non_coord):
        """If relative or not samples NonCoord raises."""
        match = "does not support relative and samples must"
        with pytest.raises(CoordError, match=match):
            basic_non_coord.select((1, 2), relative=True)
        with pytest.raises(CoordError, match=match):
            basic_non_coord.select((1, 2), samples=False)

    def test_select_samples_tuple_and_slice(self, basic_non_coord):
        """Ensure selecting with samples using tuple or slice works."""
        out1, inds1 = basic_non_coord.select((1, 5), samples=True)
        out2, inds2 = basic_non_coord.select(slice(1, 5), samples=True)
        assert inds1 == inds2
        assert out1 == out2
        assert len(out1) == 4

    def test_select_array(self, basic_non_coord):
        """Ensure an array can be used for selections."""
        ar = np.array([1, 2, 4])
        out, _ = basic_non_coord.select(ar, samples=True)
        assert isinstance(out, basic_non_coord.__class__)
        assert len(out) == len(ar)

    def test_nan_float_dtype(self):
        """Ensure data are NaN and float dtype."""
        default = get_coord(shape=10)
        data = default.data
        assert np.all(pd.isnull(data))
        assert np.issubdtype(data.dtype, np.floating)

    def test_nat_datetime_dtype(self):
        """Ensure data are Nat and datetime dtype."""
        coord = get_coord(shape=10, dtype="datetime64[us]")
        data = coord.data
        assert np.all(pd.isnull(data))
        assert np.issubdtype(data.dtype, np.datetime64)

    def test_bad_order_params_raise(self, basic_non_coord):
        """Ensure non coordinates cannot be ordered by value."""
        match = "does not support relative and samples must"
        with pytest.raises(CoordError, match=match):
            basic_non_coord.order((1, 2), relative=True)
        with pytest.raises(CoordError, match=match):
            basic_non_coord.order((1, 2), samples=False)

    def test_order_by_samples(self, basic_non_coord):
        """Ensure coord can be ordered by samples."""
        order = [0, 1]
        out, ind = basic_non_coord.order(order, samples=True)
        assert len(order) == len(out)

    def test_to_summary(self, basic_non_coord):
        """Ensure we can convert non coord to summary."""
        summary = basic_non_coord.to_summary()
        assert isinstance(summary, CoordSummary)

    def test_equals_to_other_coord(self, basic_non_coord):
        """Non cord should not be equal to other coord of same length."""
        other = get_coord(data=np.arange(len(basic_non_coord)))
        assert other != basic_non_coord
        assert basic_non_coord != other

    def test_min_max(self, basic_non_coord):
        """Min and max should both be NaN."""
        v_min, v_max = basic_non_coord.min(), basic_non_coord.max()
        assert pd.isnull(v_max)
        assert pd.isnull(v_min)

    def test_sort(self, basic_non_coord):
        """Sort should do nothing."""
        out, ind = basic_non_coord.sort()
        assert out is basic_non_coord
        assert ind == slice(None)

    def test_get_coord_array_of_nan(self):
        """An array of NaN should return uncoord."""
        array = np.empty(10) * np.nan
        out = get_coord(data=array)
        assert isinstance(out, CoordPartial)

    def test_start_stop(self):
        """Ensure start/stop/step are maintained in NonCoords."""
        coord = get_coord(shape=10, step=1)
        assert isinstance(coord, CoordPartial)
        assert coord.step == 1
        # Test start/stop
        coord = get_coord(start=10, shape=10)
        assert coord.start == 10
        assert len(coord) == 10

    def test_length_zero(self):
        """Ensure length 0 non coord is possible."""
        coord = get_coord(shape=0)
        assert coord.degenerate

    def test_uptype_with_data(self):
        """Ensure adding a data array converts to Coord Array."""
        coord = get_coord(shape=10)
        data = random_state.rand(10)
        out = coord.update(data=data)
        assert np.all(out.data == data)

    def test_uptype_with_step(self):
        """Ensure adding a step to existing length/start creates CoordRange."""
        partial = get_coord(shape=10, start=0)
        expected = get_coord(start=0, stop=10, step=1)
        out = partial.update(step=1)
        assert out == expected

    def test_init_degen(self):
        """Ensure degen is inited by any sort of empty array."""
        arrays = [
            np.empty((0,), dtype=np.float64),
            np.empty((0,), dtype=np.int_),
            np.empty((0, 10), dtype="datetime64[ns]"),
        ]
        for ar in arrays:
            out = get_coord(data=ar)
            assert out.degenerate
            assert out.dtype == ar.dtype
            assert len(out) == 0

    def test_empty(self, degenerate_time_coord):
        """Ensure empty changes length"""
        empty = degenerate_time_coord.empty()
        assert len(empty) == 0

    def test_min_max_degen(self, degenerate_time_coord):
        """Ensure min/max are nullish."""
        assert pd.isnull(degenerate_time_coord.min())
        assert pd.isnull(degenerate_time_coord.max())

    def test_degenerate_with_step_from_array(self):
        """CoordRange should be possible empty."""
        ar = np.empty(0, dtype=int)
        coord = get_coord(data=ar, step=1)
        assert coord.step == 1
        assert coord.dtype == ar.dtype

    def test_properties_degenerate_from_coord_range(self, evenly_sampled_coord):
        """Test key properties for degenerate coord."""
        coord = evenly_sampled_coord.empty()
        assert not coord.evenly_sampled
        assert not coord.sorted
        assert not coord.reverse_sorted
        assert coord.degenerate

    def test_timedelta_nulls(self):
        """Ensure timedelta coordinate returns proper null values in array."""
        coord = get_coord(shape=(10,), dtype="timedelta64[s]")
        array = coord.values
        assert np.issubdtype(array.dtype, np.timedelta64)
        assert pd.isnull(array).all()

    def test_update_degenerate_datetime_to_coord_range(self, degenerate_time_coord):
        """Ensure updating partial with enough info creates coord range."""
        out = degenerate_time_coord.update(step=dc.to_timedelta64(10)).update(
            start=dc.to_datetime64("2020-10-01")
        )
        assert isinstance(out, CoordRange)


class TestCoercion:
    """Some data types should support coercion in selecting (eg dates)."""

    def test_date_str(self, evenly_sampled_date_coord):
        """Ensure date strings get coerced."""
        drange = ("1970-01-01T00:00:01", 10)
        out, indexer = evenly_sampled_date_coord.select(drange)
        assert isinstance(out, evenly_sampled_date_coord.__class__)
        assert out.dtype == evenly_sampled_date_coord.dtype

    def test_time_delta(self, evenly_sampled_time_delta_coord):
        """Ensure date strings get coerced."""
        coord = evenly_sampled_time_delta_coord
        drange = (10, 2_000)
        out, indexer = coord.select(drange)
        assert isinstance(out, coord.__class__)
        assert out.dtype == coord.dtype


class TestGetSampleCount:
    """Tests for getting the number of samples from a coordinate."""

    def test_uneven_raises(self, monotonic_float_coord):
        """Unevenly sampled counts should raise."""
        msg = "Coordinate is not evenly sampled"
        with pytest.raises(CoordError, match=msg):
            monotonic_float_coord.get_sample_count(10)

    def test_dt_lt_step(self, evenly_sampled_coord):
        """Ensure sample count is 1 when value < dt."""
        dt = evenly_sampled_coord.step * 0.1
        out = evenly_sampled_coord.get_sample_count(dt)
        assert out == 1

    def test_datetime(self, evenly_sampled_date_coord):
        """Ensure evenly sampled date coords work."""
        dt = evenly_sampled_date_coord.step
        out = evenly_sampled_date_coord.get_sample_count(12 * dt)
        assert out == 12

    def test_timedelta(self, evenly_sampled_time_delta_coord):
        """Ensure timedelta works."""
        dt = evenly_sampled_time_delta_coord.step
        out = evenly_sampled_time_delta_coord.get_sample_count(12 * dt)
        assert out == 12

    @pytest.mark.parametrize("sample", (0, 10, 100, 42, 13))
    def test_samples(self, evenly_sampled_coord, sample):
        """Ensure value is returned when samples==True."""
        assert len(evenly_sampled_coord) >= sample
        out = evenly_sampled_coord.get_sample_count(sample, samples=True)
        assert out == sample

    def test_overask_raises(self, evenly_sampled_coord):
        """Asking for more samples than length of coordinate should raise."""
        max_len = len(evenly_sampled_coord)
        duration = evenly_sampled_coord.max() - evenly_sampled_coord.min()
        step = evenly_sampled_coord.step
        # test for using samples kwargWindow or step size is larger than
        msg = "results in a window larger than coordinate"
        with pytest.raises(ParameterError, match=msg):
            evenly_sampled_coord.get_sample_count(
                max_len + 2, samples=True, enforce_lt_coord=True
            )
        # test for using normal mode
        with pytest.raises(ParameterError, match=msg):
            evenly_sampled_coord.get_sample_count(
                duration + 2 * step, enforce_lt_coord=True
            )

    def test_non_int_raises_with_samples(self, evenly_sampled_coord):
        """Non integer values should raise when sample=True."""
        match = "must be integers"
        with pytest.raises(ParameterError, match=match):
            evenly_sampled_coord.get_sample_count(1.2, samples=True)
        with pytest.raises(ParameterError, match=match):
            evenly_sampled_coord.get_sample_count("bob", samples=True)


class TestGetNextIndex:
    """Tests for finding the next index after a value."""

    relative_test_fixture_names: ClassVar = [
        "evenly_sampled_coord",
        "evenly_sampled_date_coord",
        "evenly_sampled_time_delta_coord",
        "monotonic_float_coord",
    ]

    def test_samples(self, evenly_sampled_coord):
        """Ensure samples argument allows samples to be passed in."""
        coord = evenly_sampled_coord
        assert coord.get_next_index(0, samples=True) == 0
        assert coord.get_next_index(-1.0, samples=True) == len(coord) - 1
        coord_len = len(coord)
        with pytest.raises(ValueError, match="is out of bounds"):
            coord.get_next_index(coord_len, samples=True)

    def test_samples_time(self, evenly_sampled_date_coord):
        """Ensure samples argument allows samples to be passed in."""
        coord = evenly_sampled_date_coord
        coord.get_next_index(0, samples=True)
        assert coord.get_next_index(0, samples=True) == 0
        assert coord.get_next_index(-1.0, samples=True) == len(coord) - 1

    def test_out_of_bounds_raises(self, evenly_sampled_coord):
        """Tests for out of bounds values raising exceptions."""
        msg = "is out of bounds"
        coord = evenly_sampled_coord
        step = coord.step
        max_value = coord.max()
        min_value = coord.min()
        # ensure out of bounds samples raises
        with pytest.raises(ValueError, match=msg):
            coord.get_next_index(len(coord) + 10, samples=True)
        with pytest.raises(ValueError, match=msg):
            coord.get_next_index(-len(coord) - 10, samples=True)
        # ensure out of bounds values raises
        with pytest.raises(ValueError, match=msg):
            coord.get_next_index(min_value - step)
        with pytest.raises(ValueError, match=msg):
            coord.get_next_index(max_value + step)

    def test_out_of_bounds_suppressed(self, evenly_sampled_coord):
        """Tests for out of bounds values returning when specified."""
        coord = evenly_sampled_coord
        step = coord.step
        max_value = coord.max()
        min_value = coord.min()
        func = partial(coord.get_next_index, allow_out_of_bounds=True)
        assert func(len(coord) + 10, samples=True) == len(coord) - 1
        assert func(-len(coord) - 10, samples=True) == 0
        assert func(min_value - step) == min_value
        assert func(max_value + step) == max_value

    def test_exact_values(self, evenly_sampled_coord):
        """Ensure using exact values contained in coord return index."""
        coord = evenly_sampled_coord
        for ind in [10, 1, 5]:
            assert coord.get_next_index(ind) == ind

    def test_between_values(self, evenly_sampled_coord):
        """Ensure values between coordinate values also work."""
        coord = evenly_sampled_coord
        val1, step = coord[1], coord.step
        ind1 = evenly_sampled_coord.get_next_index(val1 + step / 4)
        ind2 = evenly_sampled_coord.get_next_index(val1 + step * 3 / 4)
        assert ind1 == ind2 == 2

    def test_units(self, evenly_sampled_float_coord_with_units):
        """Ensure values with units work."""
        coord = evenly_sampled_float_coord_with_units
        val1 = np.array([10, 20]) * get_quantity("m")
        val2 = val1.to(get_quantity("ft"))
        ind1 = coord.get_next_index(val1)
        ind2 = coord.get_next_index(val2)
        assert len(val1) == len(ind1)
        assert np.all(ind1 == ind2)

    def test_unit_array(self, evenly_sampled_float_coord_with_units):
        """Ensure an array of units returns inds of equal shape."""
        coord = evenly_sampled_float_coord_with_units.set_units("m")
        val1 = 10 * get_quantity("m")
        val2 = val1.to(get_quantity("ft"))
        ind1 = coord.get_next_index(val1)
        ind2 = coord.get_next_index(val2)
        assert ind1 == ind2

    def test_array_input(self, coord):
        """Get next coord should work with an array."""
        inds = np.array([1, 2, 3, 4])
        try:
            out1 = coord.get_next_index(inds, samples=True)
        except CoordError:
            pytest.skip(f"{coord} doesn't support get_next_index.")
        values = coord.values[inds]
        out2 = coord.get_next_index(values)
        assert np.all(out1 == inds) and np.all(out2 == out1)

    def test_out_of_bounds_array_range_coord(self, evenly_sampled_coord):
        """Ensure get index works with array that has out of bounds values."""
        coord = evenly_sampled_coord
        values = coord.values
        inputs = np.arange(3) + np.max(values)
        out = coord.get_next_index(inputs, allow_out_of_bounds=True)
        assert np.allclose(out, 99)

    def test_non_sorted_coord_raises(self, random_coord):
        """Ensure a non-sorted coord raises."""
        msg = "Coords must be sorted"
        with pytest.raises(CoordError, match=msg):
            random_coord.get_next_index(10)

    @pytest.mark.parametrize("fixture_name", relative_test_fixture_names)
    def test_relative(self, request, fixture_name):
        """Ensure relative selects work."""
        coord = request.getfixturevalue(fixture_name)
        values = coord.values
        start_val, end_val = coord.min(), coord.max()
        test_index = 10
        val = values[test_index] - start_val
        # positive case
        pos_out = coord.get_next_index(val, relative=True)
        assert pos_out == test_index
        # negative case
        val = values[-test_index] - end_val
        neg_out = coord.get_next_index(val, relative=True)
        assert neg_out == len(coord) - test_index


class TestUpdate:
    """Tests for updating coordinates."""

    def test_update_changes_length(self, evenly_sampled_coord):
        """Ensure the length can be changed."""
        coord = evenly_sampled_coord
        start, stop, step = coord.start, coord.stop, coord.step
        new = coord.update(start=start, stop=stop + step * 10)
        assert len(new) == (len(coord) + 10)

    def test_update_data(self, coord):
        """Test updating data in the coordinate."""
        if coord.degenerate:
            return
        data = coord.data
        d_range = np.max(data) - np.min(data)
        new_data = data + d_range
        new_coord = coord.update(data=new_data)
        assert isinstance(new_coord, BaseCoord)
        assert all_close(new_coord.data, new_data)
        # using the keyword 'values' is deprecated but should also work.
        new_coord = coord.update_data(values=new_data)
        assert all_close(new_coord.data, new_data)


class TestAlignTo:
    """Tests for aligning two coords."""

    def test_perfect_overlap(self, evenly_sampled_coord):
        """Ensure nothing changes with perfect overlap."""
        c1, c2, s1, s2 = evenly_sampled_coord.align_to(evenly_sampled_coord)
        assert c1 == c2 == evenly_sampled_coord
        assert s1 == s2 == slice(None, None)

    def test_subset(self, evenly_sampled_coord):
        """Ensure a subset is properly aligned."""
        coord = evenly_sampled_coord
        sub, sub_slice = coord.select((0, 10), samples=True)
        c1, c2, s1, s2 = coord.align_to(sub)
        assert c1 == c2 == sub
        assert s1 == sub_slice
        assert s2 == slice(None)

    def test_intersection(self, evenly_sampled_coord):
        """Test for when two coords overlap but aren't contained in the other."""
        coord1 = evenly_sampled_coord
        mid_ind = len(coord1) // 2
        coord2 = coord1.update_limits(min=coord1.data[mid_ind])
        c1, c2, s1, s2 = coord1.align_to(coord2)
        assert c1 == c2
        values1 = coord1.data[s1]
        values2 = coord2.data[s2]
        assert np.all(values1 == values2)

    def test_no_overlap(self, evenly_sampled_coord):
        """Ensure no overlap returns empty coord."""
        coord1 = evenly_sampled_coord
        coord2 = evenly_sampled_coord.update_limits(min=coord1.max() + 10)
        c1, c2, s1, s2 = coord1.align_to(coord2)
        assert c1 == c2
        assert len(s1) == len(s2) == len(c1) == 0

    def test_non_coord_compatible(self, evenly_sampled_coord):
        """Ensure when non_coords are compatible original coords returned."""
        coord = evenly_sampled_coord
        non_coord = get_coord(data=1)
        c1, c2, s1, s2 = coord.align_to(non_coord)
        assert c1 == coord
        c1, c2, s1, s2 = non_coord.align_to(coord)
        assert c1 == non_coord

    def test_incompatible_non_coords(self, basic_non_coord):
        """Ensure 2 incompatible non coords raise when aligning."""
        first = basic_non_coord
        second = basic_non_coord.change_length(len(first) + 3)
        msg = "Non coordinates must be the same length"
        with pytest.raises(CoordError, match=msg):
            first.align_to(second)


class TestChangeLength:
    """Tests for changing length of coordinate."""

    def test_change_length_no_change(self, evenly_sampled_coord):
        """Ensure change_length works when no length change is needed."""
        new = evenly_sampled_coord.change_length(len(evenly_sampled_coord))
        assert new == evenly_sampled_coord

    def test_change_length_lengthen(self, evenly_sampled_float_coord_with_units):
        """Ensure the length can increase."""
        coord = evenly_sampled_float_coord_with_units
        current = len(coord)
        new = coord.change_length(current + 1)
        assert len(new) == current + 1

    def test_change_length_shorten(self, evenly_sampled_float_coord_with_units):
        """Ensure the length can increase."""
        coord = evenly_sampled_float_coord_with_units
        current = len(coord)
        new = coord.change_length(current - 1)
        assert len(new) == current - 1

    def test_non_coord_change_length(self, basic_non_coord):
        """Ensure non coord can change length."""
        out = basic_non_coord.change_length(2 * len(basic_non_coord))
        assert len(out) == 2 * len(basic_non_coord)
        # The type shouldn't change.
        assert isinstance(out, basic_non_coord.__class__)

    def test_not_implemented_in_baseclass(self, evenly_sampled_coord):
        """Ensure Change coords is not implemented in base class."""
        coord = evenly_sampled_coord
        with pytest.raises(NotImplementedError):
            BaseCoord.change_length(coord, 10)


class TestIssues:
    """Tests for special issues related to coords."""

    @pytest.fixture(scope="class")
    def awkward_off_by_one(self):
        """Awkward array like one found in the wild which causes snap issues."""
        # Awkward, but this is a simplification of a time array found in the wild.
        # Unfortunately, these arrays have to be this large to trigger the off
        # by one error we are testing.
        dt1 = np.timedelta64(131457, "ns")
        dt2 = np.timedelta64(131456, "ns")

        t1 = dc.to_datetime64("2023-01-01T00:12:13.232123222")
        ar1 = t1 + np.arange(26926) * dt1

        t2 = t1.max() + 2.232323 * dt1
        ar2 = t2 + np.arange(45642 - len(ar1)) * dt2

        array = np.concatenate([ar1, ar2], axis=0)
        coord = dc.core.get_coord(data=array)
        return coord

    def test_event_1_gives_coord_range(self, event_patch_1):
        """Ensure patch1 gives the coord range."""
        time_vals = event_patch_1.coords.get_array("time")
        coord = get_coord(data=time_vals)
        assert coord.evenly_sampled

    def test_len_1_list(self):
        """A len 1 list should work as input to get_coords."""
        coord = get_coord(data=[0])
        assert isinstance(coord, BaseCoord)
        assert len(coord) == 1

    def test_update_degenerate_from_attrs(self, degenerate_time_coord):
        """Ensure updating dt on degenerate time coord doesn't fail."""
        # before this would cause infinite recursion.
        out = degenerate_time_coord.empty().update(step=10)
        assert isinstance(out, BaseCoord)
        assert len(out) == 0

    def test_snap_awkward(self, awkward_off_by_one):
        """Ensure off by one error don't occur with snap."""
        out = awkward_off_by_one.snap()
        assert out.shape == awkward_off_by_one.shape

    def test_dtype_start_int_others_float(self):
        """
        Ensure the correct dtype is given when the start value is an int
        but the other values are not.
        """
        coord = get_coord(
            start=0,
            step=1.1,
            stop=11.2,
            units="m",
        )
        assert np.issubdtype(coord.dtype, np.floating)

    def test_array_units_preserved(self):
        """Ensure the units attached to an array are preserved."""
        ft = dc.get_quantity("ft")
        array = np.array([1, 2, 3])
        out = get_coord(data=array * ft)
        assert dc.get_quantity(out.units) == ft



"""Configuration for all vizualization tests."""

from __future__ import annotations

import matplotlib.pyplot as plt
import pytest


@pytest.fixture(scope="function", autouse=True)
def close_figures():
    """Close all figures after each test."""
    yield
    plt.close("all")



"""
Core modules for reading Neubrex data.
"""

from __future__ import annotations

import numpy as np

import dascore as dc
import dascore.io.neubrex.utils_das as das_utils
import dascore.io.neubrex.utils_rfs as rfs_utils
from dascore.constants import SpoolType
from dascore.io import FiberIO
from dascore.utils.hdf5 import H5Reader


class NeubrexRFSPatchAttrs(dc.PatchAttrs):
    """Patch attrs for Neubrex files."""

    api: str | None = None
    filed_name: str = ""
    well_id: str = ""
    well_name: str = ""
    well_bore_id: str = ""


class NeubrexDASPatchAttrs(dc.PatchAttrs):
    """Patch attrs for Neubrex DAS Format files."""

    gauge_length: float = 0
    gauge_length_units: str = ""
    index_of_reflection: float = 1.46
    triggered_time: np.datetime64 | None = None
    phase_to_strain: float | None = None
    instrument_model: str = ""
    distance_decimation_filter: int = 0
    time_decimation_filter: int = 0


class NeubrexRFSV1(FiberIO):
    """
    Support for Neubrex Rayleigh Frequency Shift (DSS/DTS) version 1.

    This specifically supports DTS/DSS files recorded at the Forge cite.
    See #411.
    """

    name = "NeubrexRFS"
    preferred_extensions = ("hdf5", "h5")
    version = "1"

    def get_format(self, resource: H5Reader, **kwargs) -> tuple[str, str] | bool:
        """Determine if the resource belongs to this format."""
        if rfs_utils._is_neubrex(resource):
            return self.name, self.version
        return False

    def read(self, resource: H5Reader, snap=True, **kwargs) -> SpoolType:
        """
        Read a resource belonging to this format.

        Parameters
        ----------
        resource
            The open h5 object.
        snap
            If True, snap each coordinate to be evenly sampled.
        **kwargs
            Passed to filtering coordinates.
        """
        attr_dict, cm, data = rfs_utils._get_attrs_coords_and_data(resource, snap)
        if kwargs:
            cm, data = rfs_utils._maybe_trim_data(cm, data, **kwargs)
        attrs = NeubrexRFSPatchAttrs(**attr_dict)
        patch = dc.Patch(coords=cm, data=data[:], attrs=attrs)
        return dc.spool([patch])

    def scan(self, resource: H5Reader, snap=True, **kwargs) -> list[dc.PatchAttrs]:
        """Get the attributes of a resource belong to this type."""
        attrs, cm, data = rfs_utils._get_attrs_coords_and_data(resource, snap)
        attrs["coords"] = cm.to_summary_dict()
        attrs["path"] = resource.filename
        attrs["file_format"] = self.name
        attrs["file_version"] = self.version
        return [dc.PatchAttrs(**attrs)]


class NeubrexDASV1(FiberIO):
    """
    Support for Neubrex DAS files.
    """

    name = "NeubrexDAS"
    preferred_extensions = ("hdf5", "h5")
    version = "1"

    def get_format(self, resource: H5Reader, **kwargs) -> tuple[str, str] | bool:
        """Determine if resource belongs to this format."""
        if das_utils._is_neubrex(resource):
            return self.name, self.version
        return False

    def read(self, resource: H5Reader, **kwargs) -> SpoolType:
        """
        Read a resource of this format.

        Parameters
        ----------
        resource
            The open h5 object.
        snap
            If True, snap each coordinate to be evenly sampled.
        **kwargs
            Passed to filtering coordinates.
        """
        attr_dict, cm, data = das_utils._get_attrs_coords_and_data(resource)
        if kwargs:
            cm, data = das_utils._maybe_trim_data(cm, data, **kwargs)
        attrs = NeubrexRFSPatchAttrs(**attr_dict)
        patch = dc.Patch(coords=cm, data=data[:], attrs=attrs)
        return dc.spool([patch])

    def scan(self, resource: H5Reader, **kwargs) -> list[dc.PatchAttrs]:
        """Get the attributes of this format from File."""
        attrs, cm, data = das_utils._get_attrs_coords_and_data(resource)
        attrs["coords"] = cm.to_summary_dict()
        attrs["path"] = resource.filename
        attrs["file_format"] = self.name
        attrs["file_version"] = self.version
        return [dc.PatchAttrs(**attrs)]



"""
Utils for working with attributes.
"""

from __future__ import annotations

import warnings
from collections import ChainMap, defaultdict
from collections.abc import Sequence
from functools import reduce
from typing import Literal

import pandas as pd

import dascore as dc
from dascore.constants import attr_conflict_description
from dascore.exceptions import AttributeMergeError
from dascore.utils.docs import compose_docstring
from dascore.utils.misc import (
    _dict_list_diffs,
    all_diffs_close_enough,
    get_middle_value,
    is_valid_coord_str,
    iterate,
)


@compose_docstring(conflict_desc=attr_conflict_description)
def combine_patch_attrs(
    model_list: Sequence[dc.PatchAttrs],
    coord_name: str | None = None,
    conflicts: Literal["drop", "raise", "keep_first"] = "raise",
    drop_attrs: Sequence[str] | None = None,
    coord: None | dc.core.coords.BaseCoord = None,
) -> dc.PatchAttrs:
    """
    Merge Patch Attributes along a dimension.

    Parameters
    ----------
    model_list
        A list of models.
    coord_name
        The coordinate, usually a dimension coord, along which to merge.
    conflicts
        {conflict_desc}
    drop_attrs
        If provided, attributes which should be dropped.
    coord
        The coordinate for the new values of dim. This is provided as a
        shortcut if it has already been computed.
    """
    # TODO this is a monstrosity! need to refactor.
    model_fields = dc.core.CoordSummary.model_fields
    eq_coord_fields = set(model_fields) - {"min", "max", "step"}

    def _get_merge_coords(model_dicts):
        """Get the coordinates to merge together."""
        # pop out coord to merge on
        merge_coords = []
        for mod in model_dicts:
            coords = mod.get("coords", {})
            maybe_coord = coords.pop(coord_name, None)
            if maybe_coord is not None:
                merge_coords.append(maybe_coord)
        # dealing with empty coords
        if not merge_coords:
            return {}
        return merge_coords

    def _get_merge_dict_list(merge_coords):
        """Get a list of {attrs: []}."""
        out = defaultdict(list)
        for mdict in merge_coords:
            for key, val in mdict.items():
                out[key].append(val)
        return out

    def _get_new_coord(model_dicts):
        """Get the merged coord to set on all models."""
        merge_coords = _get_merge_coords(model_dicts)
        merge_dict_list = _get_merge_dict_list(merge_coords)
        out = {}
        # empy dicts
        if not merge_dict_list:
            return dict(merge_dict_list)
        # all these should be equal else raise.
        for key in eq_coord_fields:
            if not len(vals := merge_dict_list[key]):
                continue
            if not all(vals[0] == x for x in vals):
                msg = f"Cant merge patch attrs, key {key} not equal."
                raise AttributeMergeError(msg)
            out[key] = vals[0]
        out["min"] = min(merge_dict_list["min"])
        out["max"] = max(merge_dict_list["max"])
        step = None
        if all_diffs_close_enough(steps := merge_dict_list["step"]):
            step = get_middle_value(steps)
        out["step"] = step
        return out

    def _get_model_dict_list(mod_list):
        """Get list of model_dict, merge along dim if specified."""
        model_dicts = [
            x.model_dump(exclude_defaults=True) if not isinstance(x, dict) else x
            for x in mod_list
        ]
        # drop attributes specified.
        if drop := set(iterate(drop_attrs)):
            model_dicts = [
                {i: v for i, v in x.items() if i not in drop} for x in model_dicts
            ]
        # no coordinate to merge on, just return dicts.
        if not coord_name:
            return model_dicts
        # drop models which don't have required coords
        model_dicts = [x for x in model_dicts if coord_name in x.get("coords", {})]
        # coordinate can be determined from existing coords.
        if coord is None:
            new_coord = _get_new_coord(model_dicts)
        else:  # or if one was passed just use its summary.
            new_coord = coord.to_summary()
        if new_coord:
            for mod in model_dicts:
                mod["coords"][coord_name] = new_coord
        return model_dicts

    def _replace_null_with_none(mod_dict_list):
        """Because NaN != NaN we need to replace those values so == works."""
        out = []
        for mod in mod_dict_list:
            out.append(
                {
                    i: (v if (isinstance(v, Sequence) or not pd.isnull(v)) else None)
                    for i, v in mod.items()
                }
            )
        return out

    def _keep_eq(d1, d2):
        """Keep only the values that are equal between d1/d2."""
        out = {}
        for i in set(d1) & set(d2):
            if not d1[i] == d2[i]:
                continue
            out[i] = d1[i]
        return out

    def _handle_other_attrs(mod_dict_list):
        """Check the other attributes and handle based on conflicts param."""
        if conflicts == "keep_first":
            return [dict(ChainMap(*mod_dict_list))]
        no_null_ = _replace_null_with_none(mod_dict_list)
        all_eq = all(no_null_[0] == x for x in no_null_[1:])
        if all_eq:
            return mod_dict_list
        if conflicts == "raise":
            # determine which keys are not equal to help debug.
            uneq_keys = _dict_list_diffs(mod_dict_list)
            msg = (
                "Cannot merge models, the following non-dim attrs are not "
                f"equal: {uneq_keys}. Consider setting the `conflict` or "
                f"`attr_conflict` arguments for more flexibility in merging "
                f"unequal coordinates."
            )
            raise AttributeMergeError(msg)
        final_dict = reduce(_keep_eq, mod_dict_list)
        return [final_dict]

    mod_dict_list = _get_model_dict_list(model_list)
    mod_dict_list = _handle_other_attrs(mod_dict_list)
    first_class = model_list[0].__class__
    cls = first_class if not first_class == dict else dc.PatchAttrs
    if not len(mod_dict_list):
        msg = "Failed to combine patch attrs"
        raise AttributeMergeError(msg)
    return cls(**mod_dict_list[0])


def decompose_attrs(attr_list: Sequence[dc.PatchAttrs], exclude=("history",)):
    """Function to decompose attributes into series."""

    def _get_uri_and_hash(model):
        """Pop out this models uri and its hash."""
        uri_key = "uri" if "uri" in model else "path"
        assert uri_key in model, "all models must have uri or path"
        uri = model.pop(uri_key)
        uri_hash = hash(uri)
        return uri, uri_hash

    def _add_coords(coord_dict, dim_dict, coords, uri_hash, dims):
        """Add coordinates to structure."""
        for name, coord in coords.items():
            out_dct = coord_dict if name not in dims else dim_dict
            coord["index"] = uri_hash
            coord["units"] = "" if coord["units"] is None else coord["units"]
            # strip off the [] in datetimes/timedeltas. We will ensure the
            # precision is always ns downstream.
            if dtype := coord.get("dtype"):
                coord["dtype"] = dtype.split("[")[0]
            out_dct[name][coord["dtype"]].append(coord)

    def _add_attrs(attr_dict, model, uri_hash):
        """Add attrs to coordinate structure."""
        for name, value in model.items():
            if name.startswith("_"):
                continue
            entry = {"value": value, "index": uri_hash}
            attr_dict[name][str(type(value))].append(entry)

    def _to_df(dict_list):
        """Convert a dict list to a dataframe."""
        df = pd.DataFrame(dict_list).set_index("index").dropna(how="all")
        # TODO may need to do something faster here, not sure.
        empty_str = df == ""
        null = pd.isnull(df)
        to_keep = ~(empty_str | null).all(axis="columns")
        return df[to_keep]

    def _pandify(decom):
        """Convert all the nested dicts to series."""
        out = {}
        for key, value in decom.items():
            if isinstance(value, list):
                df = _to_df(value)
                if not df.empty:
                    out[key] = df
            else:
                out[key] = _pandify(value)
        return out

    dumped = [x.model_dump(exclude=exclude) for x in attr_list]
    out = {
        "uri": [],
        "attrs": defaultdict(lambda: defaultdict(list)),
        "coords": defaultdict(lambda: defaultdict(list)),
        "dims": defaultdict(lambda: defaultdict(list)),
    }
    for ind, model in enumerate(dumped):
        uri, uri_hash = _get_uri_and_hash(model)
        out["uri"].append({"uri": uri, "index": uri_hash})
        dims = model.get("dims", "").split(",")
        # need to handle coordinates and regular attrs differently
        coords = model.pop("coords", {})
        _add_coords(out["coords"], out["dims"], coords, uri_hash, dims)
        _add_attrs(out["attrs"], model, uri_hash)
    return _pandify(out)


def separate_coord_info(
    obj,
    dims: tuple[str, ...] | None = None,
    required: Sequence[str] | None = None,
    cant_be_alone: tuple[str, ...] = ("units", "dtype"),
) -> tuple[dict, dict]:
    """
    Separate coordinate information from attr dict.

    These can be in the flat-form (ie {time_min, time_max, time_step, ...})
    or a nested coord: {coords: {time: {min, max, step}}

    Parameters
    ----------
    obj
        The object or model to
    dims
        The dimension to look for.
    required
        If provided, the required attributes (e.g., min, max, step).
    cant_be_alone
        names which cannot be on their own.

    Returns
    -------
    coord_dict and attrs_dict.
    """

    def _meets_required(coord_dict, strict=True):
        """
        Return True coord dict meets the minimum required keys.

        coord_dict represents potential coordinate fields.

        Strict ensures all required values exist.
        """
        if not coord_dict:
            return False
        if not required and (set(coord_dict) - cant_be_alone):
            return True
        if required or not strict:
            return set(coord_dict).issuperset(required)
        return False

    def _get_dims(obj):
        """Try to ascertain dims from keys in obj."""
        # check first for coord manager
        if isinstance(obj, dict) and hasattr(obj.get("coords", None), "dims"):
            return obj["coords"].dims

        # This object already has dims, just honor it.
        if dims := obj.get("dims", None):
            return tuple(dims.split(",")) if isinstance(dims, str) else dims

        potential_keys = defaultdict(set)
        for key in obj:
            if not is_valid_coord_str(key):
                continue
            potential_keys[key.split("_")[0]].add(key.split("_")[1])
        return tuple(i for i, v in potential_keys.items() if _meets_required(v))

    def _get_coords_from_top_level(obj, out, dims):
        """First get coord info from top level."""
        for dim in iterate(dims):
            potential_coord = {
                i.split("_")[1]: v for i, v in obj.items() if is_valid_coord_str(i, dim)
            }
            # nasty hack for handling d_{dim} for backward compatibility.
            if (bad_name := f"d_{dim}") in obj:
                msg = f"d_{dim} is deprecated, use {dim}_step"
                warnings.warn(msg, DeprecationWarning, stacklevel=3)
                potential_coord["step"] = obj[bad_name]

            if _meets_required(potential_coord, strict=False):
                out[dim] = potential_coord

    def _get_coords_from_coord_level(obj, out):
        """Get coords from coordinate level."""
        coords = obj.get("coords", {})
        if hasattr(coords, "to_summary_dict"):
            coords = coords.to_summary_dict()
        for key, value in coords.items():
            if hasattr(value, "to_summary"):
                value = value.to_summary()
            if hasattr(value, "model_dump"):
                value = value.model_dump()
            if _meets_required(value, strict=False):
                out[key] = value

    def _pop_keys(obj, out):
        """Pop out old keys for attrs, and unused keys from out."""
        # first coord subdict
        obj.pop("coords", None)
        # then top-level
        for coord_name, sub_dict in out.items():
            for thing_name in sub_dict:
                obj.pop(f"{coord_name}_{thing_name}", None)
            if "step" in sub_dict:
                obj.pop(f"d_{coord_name}", None)

    # sequence of short-circuit checks
    coord_dict = {}
    required = set(required) if required is not None else set()
    cant_be_alone = set(cant_be_alone)
    if obj is None:
        return coord_dict, {}
    if hasattr(obj, "model_dump"):
        obj = obj.model_dump()
    obj = dict(obj)
    # Check if dims need to be updated.
    new_dims = _get_dims(obj)
    if new_dims and new_dims != dims:
        obj["dims"] = new_dims
        dims = new_dims
    # this is already a dict of coord info.
    if dims and set(dims).issubset(set(obj)):
        return obj, {}
    _get_coords_from_coord_level(obj, coord_dict)
    _get_coords_from_top_level(obj, coord_dict, dims)
    _pop_keys(obj, coord_dict)
    if "dims" not in obj and dims is not None:
        obj["dims"] = dims
    return coord_dict, obj



"""
Tests specific to the Sentek format.
"""

import numpy as np

from dascore.compat import random_state
from dascore.io.sentek import SentekV5


class TestSentekV5:
    """Tests for Sentek format that aren;t covered by common tests."""

    def test_das_extension_not_sentek(self, tmp_path_factory):
        """Ensure a non-sentek file with a das extension isn't id as sentek."""
        path = tmp_path_factory.mktemp("sentek_test") / "not_sentek.das"
        ar = random_state.random(10)
        with path.open("wb") as fi:
            np.save(fi, ar)
        sentek = SentekV5()
        assert not sentek.get_format(path)



"""
Module for reading DAS data recorded by Sentek interrogator

Examples
--------

import dascore as dc
from dascore.utils.downloader import fetch

path_to_sentek_file = fetch("DASDMSShot00_20230328155653619.das")
sentek_patch = dc.spool(path_to_sentek_file)[0]
"""
from .core import SentekV5
